//                              -*- Mode: C++ -*-
//: EDClusTraHypotModifier.C
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 2/5/97
// Last Update ........
//

#include "care/care.hh"
#include "care/EMClusterTree.hh"
#include "care/EDClusTraHypotModifier.hh"
#include "care/ECBuilder.hh"
#include "arte/ArteMessage.hh"
#undef WRITE

/* 
CLASS
  EDClusTraHypotModifier
  Class for finding the number of hypothetical particles gave origin to
  a cluster.

  The algorithm choseen make use of a constant CutOff level to define an
  hypotetical cluster.


*/


//////////////// EDClusTraHypotModifier Static data Definitions //////////////

//////////////////// EDClusTraHypotModifier Class Definitions ////////////////

// 
EDClusTraHypotModifier::EDClusTraHypotModifier(ECBuilder* pbuilder) : 
  hyp(), builder(pbuilder) {}

// 
EDClusTraHypotModifier::~EDClusTraHypotModifier(){}

//
//--
//  Base function to be called with the TopCluster List
//
void EDClusTraHypotModifier::analyzeTopList(EMClusterTree & topList){
  doModify(&topList);
}

//
//--
//  Base function to be called with the TopCluster List
//
void EDClusTraHypotModifier::doModify(EMClusterTree * tree){
  ECClHyp sumHyp;
# ifdef WRITE
  arteMsg<<" Analyzing top list for new tracks";
  arteMsg.allp();
# endif
  for(EMClusterMember* child = tree->begin(); child!=tree->end(); 
      child = child->next()){
    if( !child->isALeaf()){
      hyp.reset();
#     ifdef WRITE
      EMCluster* clus = (EMCluster*) child;
      arteMsg<<" - Cluster hypotheses: ";
      clus->getAllHypotheses().dump();
#     endif
      child->modifyMe(*this); 
#     ifdef WRITE
      arteMsg<<" - Cluster hypotheses: ";
      clus->getAllHypotheses().dump();
#     endif
      sumHyp += hyp;
    }
  }
  hyp = sumHyp;
  tree->setAllHypotheses(hyp);
}


//--
// Cluster Analyzer
void EDClusTraHypotModifier::doModify(EMCluster * clus){
  //-- 
  // For each cluster initialize a new counter
  //
  if( clus->begin()->isALeaf() ){
    // get data for the current cluster
    hyp = clus->getAllHypotheses();
    
    // Analyze a Base Cluster
    // if there are tracks then
    // count them and increase the number of hypotheses
    if( hyp.getTracks() >0 ){
      //
      // in future I have to check if tracks are in the peak of cluster
      //
      hyp.setChargHyp(1); // all Tracks are hypot for charged
      hyp.setHypotheses(1); // and these is no space for anything else
    } else if( hyp.getHypotheses()>0 ){  // it should be always 1
      hyp.setNeuHyp(1); // the cluster is a neutral one
      if( hyp.getHypotheses()!=1){
	//
	// Printout a warning and count it
	//
	static int cluTH_Warn = 0xECA0004;
	arteMsg<<"Error in EAClusTraHypot: base cluster Hyp!=1: "
	    <<hyp.getHypotheses();
	arteMsg.warn(cluTH_Warn);
      }
    } else {
      // here for low energy clusters
      //  hyp.numTracks==0 and hyp.numHypot==0
      //      hyp.numVoid = 1;
    } 
  } else {
    // analyze a structured cluster
    // start analyzing subcluster
    bool freetrack = false;
    ECClHyp sumHyp;
    for(EMClusterMember* child = clus->begin(); 
	child!=clus->end();  child = child->next()){
      hyp.reset();
      child->modifyMe(*this);
      sumHyp += hyp;
      freetrack  = freetrack || (hyp.getTracks()==1 && hyp.getHypotheses()==0);
    }
    // now test if there are tracks pointing to cells in the merging region
    if( freetrack ){
      //
      // ok there are unmatched tracks to match to some cluster
      //
      // start considering Track cells
      //
      for(EMClusterMember* child = clus->begin(); 
	  child!=clus->end();  child = child->next()){
	if( child->isALeaf() && child->maxEnergy()<0.0001 ){
	  EMTrackOnEcal* track = (TrackOnEcal*) child;
	  
	  // now find where the neibouring cells of the cluster pointed are
	  EMCluster* newclus = builder->findOtherCluster(track);
	  if( newclus!=0 && newclus!=clus ){	    
	    // a void or a neutral cluster has been found 
	    //
	    // move this track to cluster "newclus"
	    //
	    child = child->previous(); // remember that the child pointer
	                               // will be no more valid after 
	                               // moving the child
	    newclus->move(track);
	    //
	    // update parameters of the parent cluster
	    //
	    Hyp_t neutr = newclus->getNeuHyp();
	    EMCluster* parent = newclus;
	    while( parent!=clus && parent!=0){
	      // for each parent cluster
	      // there is a true track in addition
	      parent->newTrack();
	      // there is a track hypothesis
	      parent->setChargHyp(parent->getChargHyp()+1);
	      // there is 'neutr' neutrals less (neutr = 0 or 1)
	      if( neutr!=0)
		parent->setNeuHyp(parent->getNeuHyp()-neutr);
	      else
		// there is a general hypotheses more
		parent-> setHypotheses(parent->getHypotheses()+1);
	      parent = parent->getParent();
	    }
	    
	    // current cluster has a charged hypotheses more
	    // and neutr hypotheses less
	    clus->setChargHyp(clus->getChargHyp()+1);
	    if( neutr!=0)
	      clus->setNeuHyp(clus->getNeuHyp()-neutr);
	    else
	      clus->setHypotheses(clus->getHypotheses()+1);
	    //
	    sumHyp.setChargHyp(sumHyp.getChargHyp()+1);
	    sumHyp.setNeuHyp(sumHyp.getNeuHyp()-neutr);
	    sumHyp.setHypotheses(sumHyp.getHypotheses()+1-neutr);
	  } else {
	    // no old cluster found where to put this track
	    // then  make a new cluster with cells of the  boundary region
	    // Take care of the possibility of more tracks that are
	    // extrapolated to a single boundary region cell
	    // 
	    // Start testing if the parent cluster of the current track 
	    // is the same of that of the cell in which the track is
	    
	    newclus = track->whichCell()->getParent();
	    if( track->getParent()==newclus){
	      // ok it is the same: then make a new cluster
	      newclus = builder->newCluster(track->whichCell()); 
	      // 
	    } else {
	      // it is not the same: then move the current track to the 
	      // parent cluster of the HitCell pointed.
	      newclus = track->whichCell()->getParent();
	    }
	    // move this track to cluster "newclus"
	    //
	    child= child->previous(); // remember that the child pointer
      	                              // will be no more valid after 
	                              // moving the child
	    newclus->move(track);
	    //
	    // update parameters of the parent cluster
	    //
	    Hyp_t neutr = newclus->getNeuHyp();
	    EMCluster* parent = newclus;
	    while( parent!=clus && parent!=0){
	      // for each parent cluster
	      // there is a true track in addition
	      parent->newTrack();
	      // there is a track hypothesis
	      parent->setChargHyp(parent->getChargHyp()+1);
	      // there is 'neutr' neutrals less (neutr = 0 or 1)
	      if( neutr!=0)
		parent->setNeuHyp(parent->getNeuHyp()-neutr);
	      else
		// there is a general hypotheses more
		parent-> setHypotheses(parent->getHypotheses()+1);
	      parent = parent->getParent();
	    }
	    // current cluster has a charged hypotheses more
	    // and neutr hypotheses less
	    clus->setChargHyp(clus->getChargHyp()+1);
	    if( neutr!=0)
	      clus->setNeuHyp(clus->getNeuHyp()-neutr);
	    else
		clus->setHypotheses(clus->getHypotheses()+1);
	    sumHyp.setChargHyp(sumHyp.getChargHyp()+1);
	    sumHyp.setNeuHyp(sumHyp.getNeuHyp()-neutr);
	    sumHyp.setHypotheses(sumHyp.getHypotheses()+1-neutr);
	  }
	}
      }
    }
    hyp = sumHyp;
  } 
  // At this point it is satisfied that
  //  hyp.numHypot =  hyp.numCharg + hyp.numNeutr 
  //
  // there are changes to the stored value of Hypotheses
  // update them
  //
  clus->setAllHypotheses(hyp);
}

//--
// HitCell Analyzer
void EDClusTraHypotModifier::doModify(EMHitCell * ){

}

//--
// HitCell Analyzer
void EDClusTraHypotModifier::doModify(EMClusterMember * ){
  
}


//--
// TrackOnEcal Analyzer
void EDClusTraHypotModifier::doModify(EMTrackOnEcal * ){
  hyp = ECClHyp(0,1,0,0); // one track
}

