/************************************************************************
* This file contains the routines which implement the cell model for
* the outer tracker digitization.
*
* All routines called from dgomod
*
* Written by S.Nowak, C.Stegmann, A.Lanyov, W. Hulsbergen  June 2000
*
************************************************************************/

/************************************************************************
* The application of the cell model in the outer tracker digitization
* exists of the following steps:
*
*  Given a MIMP and a module
*   1) calculate distance $dd$ to wire in module that are `close' to track
*   2) check whether channel exists and is good (level >=3 only)
*   3) calculate length $l$ of segment of track through cell 
*   4) generate cluster on segment closest to the distance of closest
*      approach to wire and check whether closer than $0.5*l$
*   5) calculate distance of closest cluster to wire, convert distance
*      to drifttime and apply gaussian smearing. This is the measurement.
*
*  After processing all MIMPs in a module
*   6) create hit in cell for smallest measured time in cell.
*   7) convert measured time to expectation value for distance 
*     (hitb->q) and for the sqrt(variance) of the distance (hitb->r)
*  
*  Step 7 is new with repect to older versions of the digitization. It
*  mimics the effect of the calibration as applied to real data.
*
*  Wouter Hulsbergen (June 2000)
************************************************************************/

#include "digi/dgocell.hh"
#include <math.h>
#include <iostream>
#include <arte/arte.hh>

// This is not really an example of nice C++ programming but for now
// it should just work. 

extern "C" 
{
  void grndm_(float [], const int &);
  void granor_(float &, float &);
}

namespace dgocell
{
  // default parameters of the model
  float gPrimIonDensity = 20 ;      // prim. ion. density [1/cm]
  float gAttachDensity  = 0.0 ;     // attachment density [1/cm]
  float gGausSmearing   = 0.02 ;    // gaussian smearing 
  float gNominalEfficiency = 0.98 ; // extra cut on efficiency 
  float gDriftVelocity = 0.01 ;     // [cm/ns]
  bool gIsInitialized = false ;

  void Print()
  { 
    std::cout << "**************************************************" << std::endl 
	      << "* Initializing otr digitization                  *" << std::endl
	      << "*                                                *" << std::endl
	      << "* Model parameters:                              *" << std::endl
	      << "*  primary ion density [1/cm]: " << gPrimIonDensity << std::endl
	      << "*  attachment density [1/cm]:  " << gAttachDensity << std::endl 
	      << "*  nominal efficiency : "        << gNominalEfficiency << std::endl
	      << "*  gaussian smearing : "         << gGausSmearing << std::endl
	      << "* Model output (averaged over modules):          *" << std::endl 
	      << "*  total efficiency 5 mm (at wire) : " 
	      << gNominalEfficiency * (1-exp(-gPrimIonDensity*0.5)) << std::endl
	      << "*  total efficiency 10 mm (at wire) : " 
	      << gNominalEfficiency * (1-exp(-gPrimIonDensity*1.0)) << std::endl
	      << "*  total resolution (at 1 mm) : " 
	      << EstimatedSigma(0.1/gDriftVelocity,gGausSmearing) << std::endl
	      << "**************************************************" << std::endl ;
  }
  
  void Initialize()
  {
    double acc2sum(0),effsum(0) ;
    unsigned int n(0) ;  
    for(ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin() ;
	gedeIt != ArteTable<GEDE>::end() ; ++gedeIt) 
      if(gedeIt->cmp==2 && gedeIt->sec>0) { 
	acc2sum += gedeIt->acc*gedeIt->acc ;
	effsum += gedeIt->eff ;
	n++;
      }
    gNominalEfficiency = effsum/n;
    gGausSmearing = sqrt(acc2sum/n) ;
    gIsInitialized=true ;
    Print() ;
  }

  float PathLengthHC(float pitch, float x, float tx, float ty)
  {
    // This routine calculates the path lenth of a track trough a
    // honeycomb cell.
    // pitch: the wire pitch as in gede->pitch
    // x:  (x_track - x_wire) in plane z_wire
    // tx: tx_track in system of wire (rotated)
    // ty: ty_track in system of wire (rotated)
    
    // The coordinates normalized to the pitch are called xp and zp.
    // These are the coordinates of the 6 sides, normalized to the pitch:
    static const float xp[6] = {-1/3.,-1/6.,1/6.,1/3.,1/6.,-1/6.} ;
#define REC2SQRT3 1/(2*sqrt(3))
    static const float zp[6] = {0,REC2SQRT3,REC2SQRT3,0,-REC2SQRT3,-REC2SQRT3} ;
#undef REC2SQRT3
    
    // Loop over all sides and find points of intersection
    const float xptr = x/pitch; // the xp-coordinate of the track
    float xpent[2];             // the xp-coordinates of the intersections
    int n = 0;                  // counter for the intersections
    for(int i1=0; i1<6; i1++) {
      int i2=(i1+1)%6 ;
      // This is xp of point of intersect
      float xis = ((xp[i2]-xp[i1])*xptr + tx*(xp[i2]*zp[i1] - xp[i1]*zp[i2])) /
	( (xp[i2]-xp[i1]) - tx* (zp[i2]-zp[i1]) ) ;
      // Is this between boundaries ?
      if( (xis - xp[i1])*(xp[i2] - xis) >0 ) {
	if(n<=1) xpent[n] = xis ;
	n++;
      }
    }
    
    // Calculate the path length from the intersections
    float len=0;
    if(n==2) {
      float dz = pitch * fabs((xpent[1]-xpent[0])/tx) ;
      len = dz * sqrt(1 + tx*tx + ty*ty) ;
    }
    return len;
  }
  
  float FirstCluster(float dd /* distance to wire */, float eff) 
  {
    /* This routine simulates the distance of the closest cluster to any
       point on the track. The probability for `no attachment' is a
       function of the distance to the wire. A lower estimate is given
       by exp(fabs(dd)*attachDensity) ; This results in a `lam' effective
       ionization density.  Given this effective density we use a poison
       estimate of the closest cluster to the point, following the
       probability P(dlen) = exp(-2*dlen*lam) (The factor 2 comes in
       because the `closest cluster' can be in two directions, thus
       halving the average distance.)  */
    
    int i1 = 1;
    float dlen(0), rndm[1];
    float lam = gPrimIonDensity * exp(fabs(dd)*gAttachDensity) ;
    grndm_(rndm,i1);
    dlen = -log(rndm[0])/(2*lam) ;
    
    if(eff<1) {
      grndm_(rndm,i1);
      if(rndm[0]>eff) dlen += 100 ; // this takes care hit is not valid
    }  
    return dlen;
  }
  
  float MeasuredTime( float d /* distance of first cluster */, float acc)
  {
    if(!gIsInitialized) Initialize() ;
    float rndm[2] ;
    granor_(rndm[0],rndm[1]);
    float dm = d + acc*rndm[0] ;
    return dm/gDriftVelocity ;
  }

  // The next two routines somehow mimic the effect of
  // calibration. They should take care that the average residual is
  // more or less zero and that hitb->r is a good measurement of the
  // resolution.

  float EstimatedRadius( float t /* measured time */, float acc) {
    // calculate driftdistance from drifttime
    float r = t*gDriftVelocity ;
    // if smaller than zero, truncate
    if(r<0.001) r=0.001;
    float lam = gPrimIonDensity * exp(r*gAttachDensity) ; 
    // correction for primion:
    float xi = 2*r*lam ;
    r = r/(1 + sqrt(xi)*exp(-xi)) ;
    // correction for smearing (not yet understood)
    r = sqrt(r*r+acc*acc*2/3.1415927) ;
    return r;
  }
  
  float EstimatedSigma( float t /* measured time */, float acc) {
    
    float r = t*gDriftVelocity ;
    if(r<0) r=0;
    float lam = gPrimIonDensity * exp(r*gAttachDensity) ;
    float xi = 2*r*lam ;
    float sclus = 1/(3*lam)*(1-exp(-xi))/(1 + exp(-xi)) ;
    if (sclus <0.001) sclus = 0.001 ;
    return sqrt(sclus*sclus + acc*acc) ;
  }
}  

// the kuip interface for setting parameters
extern "C" 
{  
  extern float ku_getr();

  void dgocellSetPrimIonDensity() {
    dgocell::gPrimIonDensity = ku_getr();
    dgocell::gIsInitialized=false ;
  }
  
  void dgocellSetAttachDensity() {
    dgocell::gAttachDensity = ku_getr();
    dgocell::gIsInitialized=false ;
  }
  
  void dgocellSetGausSmearing() {
    float f = ku_getr() ;
    for(ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin() ;
	gedeIt != ArteTable<GEDE>::end() ; gedeIt++) 
      if(gedeIt->cmp==2) gedeIt->acc = f ;
    dgocell::gGausSmearing = f; 
    dgocell::gIsInitialized=false ;
  }
  
  void dgocellSetNominalEfficiency() {  
    float f = ku_getr() ;
    for(ArteTable<GEDE>::iterator gedeIt = ArteTable<GEDE>::begin() ;
	gedeIt != ArteTable<GEDE>::end() ; gedeIt++) 
      if(gedeIt->cmp==2) gedeIt->eff = f ;
    dgocell::gNominalEfficiency = f; 
    dgocell::gIsInitialized=false ;
  }
}
