#include <iostream>
#include <algorithm>
#include "arte/arte.hh"
#include "grover/higz.h"
#include "grover/grover.hh"

//
// Part of the Grover package
// Please see the header file grover.hh
// for authors etc.
//
//
// Modifications:
// When?     Who?     What?
// YY-MM-DD  Name     Modification
// -----------------------------------------------------------------------------
// 14-05-00  Max      First Version   
// 24-05-00  Max      Corrected bug in target alignment
//


// Wire geometry for alignment
float wire_xpos[9]  = {0.0},
      wire_ypos[9]  = {0.0},
      wire_zpos[9]  = {0.0},
      wire_xsize[9] = {0.0},
      wire_ysize[9] = {0.0},
      wire_zsize[9] = {0.0};

// Arrays with vertex-rseg relations
int grover_rseg_to_rver[1500] = {0};

const int ROVERMAXTRACKS = 1000;

extern bool GroverDistSort(GroverParticle,
			   GroverParticle);

int GroverPrimaries(GroverParticleList* allParticles,
		    GroverWireList*     allWires,
		    GroverCuts*         cuts,
		    GroverVertexList*   allVertices) 
{
  ////////////////////////////////////////////////////////////////////
  //
  //                     GroverPrimaries
  //                               
  // Finds the primary vertices
  //
  // Authors: Dmitry Emilianov, Ivan Kisel, Max Sang, Yuri Vassiliev
  //
  // Inputs (all pointers): A particle list 
  //                        A wire list
  //                        A set of cuts
  // Output: A list of vertices in *allVertices
  //         Note that the lists of particles and wires 
  //         might also be changed during execution!
  //
  // Everything is done in VDS local coordinates!
  // Grover objects can transform themselves between
  // Global and VDS Local systems. However, the covariances
  // might not be correct after transformation!
  //
  /////////////////////////////////////////////////////////////////////


  static int EventNumber = 0;
  static bool firstcall = true;
  EventNumber++;
  firstcall = (EventNumber==1);


  int nver = 0;                // number of vertices found in this event
  GroverParticleList::iterator trackonwire[9][200]; // pointers to particles
  GroverParticleList::iterator trackincluster[200]; //  -"-
  float  trackweight[9][200];  // weights for associating tracks to wires
  int    ntronwire[9];         // number of tracks on wire [i]


  // Get cuts
  static bool  verbose           = (cuts->Verbose == 1);
  static bool  roverdraw         = (cuts->GroverDraw == 1);
  static int   mintracksinvertex = cuts->RPMinTrksInVtx;
  static float wtsig             = cuts->RPTrkWireSigmaMax; 


  // Initialisation
  if (firstcall) {
    cout << "********* Grover v."
	 << GroverVersion << ": Finding Primaries with Rover **********" << endl;
    // Initialise graphics and draw the beam spot
    // (with approximately correct offset if real data)
    if (roverdraw) graph_ini();
  }


  // push grover to reread targets for each new run
   ArtePointer<EVHD> evhd = *(ArteTable<EVHD>::begin());
   static int run_old;

   if(firstcall) 
     run_old = evhd->run;

   if (evhd->run != run_old)
     rover_newgeometry = 1;
   run_old = evhd->run;
   


  // Make a static copy of the wire list in local coords which will
  // gradually be adjusted according to the distribution 
  // of primaries.
  static GroverWireList newWires;
  if (firstcall||rover_newgeometry) {

    newWires.clear();  // clear in case of new geometry
    if (rover_newgeometry) rover_newgeometry = 0;
    //float beam_x = 0.30, beam_y = -0.05; 
    // beamspot in local coords for MC and real

    float beam_x = 0.0, beam_y = 0.0;  // beam center for simu data
    //    if (!ArteTable<MCIN>::booked()){
      beam_x =  0.30;                  // beam center for real data
      beam_y = -0.05;
      //    } 
    for (GroverWireList::iterator iwire = allWires->begin();
	 iwire != allWires->end(); ++iwire) {
      if (  (fabs(iwire->Y())<1.0) && (iwire->Horizontal())  ||
	    (fabs(iwire->X())<1.0) && (iwire->Vertical())
	    ) { // active wire
	GroverWire temp = *iwire; // is it correct ???
	temp.SetSizeX(0.1); // force the wires to be 1mm blobs, initially
	temp.SetSizeY(0.1);
	temp.SetNPrimaries(100);
	temp.SetNParticles(100);
	temp.GlobalToLocal();
	// fix coordinates to beamspot
	if (temp.Horizontal()) temp.SetX(beam_x);
	if (temp.Vertical())   temp.SetY(beam_y);
	temp.SetNParticles(100);  // weight for GTAR information
	temp.SetNPrimaries(100);
	newWires.push_back(temp); // store new wire
      }
    }
    if (verbose) cout << "*** GroverPrimaries: number of active wires = "
		      << newWires.size() << endl;
  }


  // Draw the target wires
  if (roverdraw) draw_targets(&newWires);


  // If MC data, draw the MC vertices. Also draw the MC 
  // tracks where they intersect with zplane = whatever.
  float zplane = -1.0;
  if (true && roverdraw && ArteTable<MCIN>::booked())  draw_mc(zplane);


  // Label all the particles with an index (starting at 1!)
  // In fact, as long as you never removed anything from 
  // *allParticles, you could use e.g. (*allParticles)[i].Momentum()
  // for indexed access to the particles in the list
  // (where i starts at 0, not at 1!)
  // Also check they're all in local coordinates
  int idx = 0;
  for (GroverParticleList::iterator ipar = allParticles->begin();
       ipar != allParticles->end(); ++ipar) {
    ipar->GlobalToLocal(); // has no effect if already local
    idx++;
    ipar->SetIndex(idx);
  }
  if (verbose)  cout << "*** GroverPrimaries: Number of particles = " 
		     << idx << endl; 


  // Calculate new target position using only tracks
  {
    float dx  = 0.1, dy  = 0.1; // tracks > 1mm away are not considered
    float dxt = 0.0, dyt = 0.0, xt=0.0, yt=0.0, zt=0.0;
    float x=0.0, y=0.0;
    int   nt_tr=0, nt_vx=0;
    for (GroverWireList::iterator iwire = newWires.begin();
	 iwire != newWires.end(); ++iwire) {
      // get current position of wire
      xt = iwire->X();   yt = iwire->Y();   zt = iwire->Z();
      dxt = iwire->SizeX();  dyt = iwire->SizeY();
      nt_tr = iwire->NParticles();  nt_vx = iwire->NPrimaries();
      // -> calculate new (weighted) position
      for (GroverParticleList::iterator ipar = allParticles->begin();
	   ipar != allParticles->end(); ++ipar) {
	ipar->PropagateBeginTo(iwire->Z());
	x = ipar->BeginX();   y = ipar->BeginY();
	if ((fabs(x-xt)<dx) && (fabs(y-yt)<dy)) {
	  xt = xt*(nt_tr+nt_vx)+x;
	  yt = yt*(nt_tr+nt_vx)+y;
	  nt_tr++;
	  xt/=(nt_tr+nt_vx);  
	  yt/=(nt_tr+nt_vx);

	  dxt = (dxt*(nt_tr-1+nt_vx) + fabs(x-xt)) / (nt_tr + nt_vx);
	  dyt = (dyt*(nt_tr-1+nt_vx) + fabs(y-yt)) / (nt_tr + nt_vx);

	}
      }
      // -> set new position
      iwire->SetX(xt);
      iwire->SetY(yt);
      iwire->SetSizeX(dxt);
      iwire->SetSizeY(dyt);
      // after 10,000 tracks or vertices -> loosen up, man! The target
      // moves slowly, so don't force it to stay in the same place for 
      // ever.
      if (nt_tr+nt_vx > 10000)  { nt_tr=100;  nt_vx = 100;} 
      iwire->SetNParticles(nt_tr);
      iwire->SetNPrimaries(nt_vx);
    }
  }

  // Associate tracks to wires
  int wnr, ntr;
  for(GroverWireList::iterator iwire = newWires.begin();
      iwire != newWires.end(); ++iwire) {
    wnr = iwire->Number();
    ntr = 0;
    for(GroverParticleList::iterator ipart = allParticles->begin();
	ipart != allParticles->end(); ++ipart) {
      //ipart->PropagateBeginTo(iwire->Z()); // already inside Chi2DistTo(iwire) !!!
      if(ipart->Chi2DistTo(iwire) <= (wtsig*wtsig)){
	ntr++;
	trackonwire[wnr][ntr] = ipart;
	trackweight[wnr][ntr] = 0.0;
      }
    }
    ntronwire[wnr] = ntr;
    if (verbose && false) cout << "*** GroverPrimaries: " << ntr 
      << " particles on wire " << iwire->Number() << endl;
  } 


  // just some variables
  int ntrincluster = 0;
  GroverParticleList::iterator  it_best, it, jt;
  GroverWireList::iterator  wire_best;
  float w_best = 0.;
  int max_clu  = 0;
  int tracksonwire;
  int itrnew = 0;
  wnr = 0;


  // Search for clusters to make vertices
  while(true) { // i.e. while there is at least one seed with >=3 tracks

    it_best     = allParticles->begin();
    w_best      = 0.0;
    max_clu     = 0;
    wire_best   = 0;

    for(GroverWireList::iterator iwire = newWires.begin();
	iwire != newWires.end(); ++iwire) {

      // Propagate all the tracks on this wire to its z plane
      wnr = iwire->Number();
      tracksonwire = ntronwire[wnr];
      if(tracksonwire < mintracksinvertex) continue; // skip to next wire
      itrnew = 0;
      for (int itr=1; itr <= tracksonwire; itr++){ 
	it = trackonwire[wnr][itr];
	if (it->UsedInPrimary() != 0) continue; // skip used tracks
	itrnew++;
	trackonwire[wnr][itrnew] = trackonwire[wnr][itr];
	trackweight[wnr][itrnew] = trackweight[wnr][itr];
	it->PropagateBeginTo(iwire->Z());
      }
      tracksonwire = ntronwire[wnr] = itrnew;
      if(tracksonwire < mintracksinvertex) // no three-track clusters left
	continue;                          // on this wire -> skip to next.


      // Find clusters (vertex seeds)
      float x1, y1, x2, y2, weight; 
      int ntr_clu = 0;
      for (int itr=1; itr <= tracksonwire; itr++){
	it = trackonwire[wnr][itr];
	x1 = it->BeginX();
	y1 = it->BeginY();
	ntr_clu = 0;
	weight  = 0.0;
	for (int jtr=1; jtr <= tracksonwire; jtr++){ // <- is this correct?
	  jt = trackonwire[wnr][jtr];                // yes, it gives minimum 
	  x2 = jt->BeginX();                         // weight = 1, that 
	  y2 = jt->BeginY();                         // characterises number
          float dist2=(x1-x2)*(x1-x2)+(y1-y2)*(y1-y2);// of tracks in
          float cut2 = wtsig*wtsig*                  // a cluster
            ( it->BeginCov(0)+it->BeginCov(2)        
             +jt->BeginCov(0)+jt->BeginCov(2));       
          if(dist2 < cut2){
            weight += ((1.0-(dist2/cut2))*(1.0-(dist2/cut2)));
            ntr_clu++;
          }
	}
	it->SetPrimaryWeight(weight);
	if (weight > w_best){ // keep best vtx candidate
	  w_best  = weight;   // on this wire
	  it_best = it; 
	  max_clu = ntr_clu; 
	  wire_best = iwire;
	}
      }


    } // finished loop over wires


    if (w_best == 0.0) break; // no vtx candidates left - leave the inf. loop.


    // propagate tracks in the best vtx seed to the appropriate wire
    wnr = wire_best->Number();
    tracksonwire = ntronwire[wnr];
    for (int itr=1; itr <= tracksonwire; itr++){
      it = trackonwire[wnr][itr];
      it->PropagateBeginTo(wire_best->Z());
    }
    
    // calculate track weights for this (closest) wire
    float x1, y1, x2, y2, weight;
    for (int itr=1; itr <= tracksonwire; itr++){
      it = trackonwire[wnr][itr];
      x1 = it->BeginX();
      y1 = it->BeginY();
      weight = 0.0;
      for (int jtr=1; jtr <= tracksonwire; jtr++) {
	jt = trackonwire[wnr][jtr];
	x2 = jt->BeginX();
	y2 = jt->BeginY();
        float dist2 = (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2);
        float cut2 = wtsig*wtsig*
          (it->BeginCov(0) + it->BeginCov(2) +
           jt->BeginCov(0) + jt->BeginCov(2) );
        if(dist2 < cut2)
          weight += ((1.0-(dist2/cut2))*
		     (1.0-(dist2/cut2)) );
      }
      it->SetPrimaryWeight(weight);
    }


    float x_best  = it_best->BeginX();
    float y_best  = it_best->BeginY();
    float c0_best = it_best->BeginCov(0);
    float c2_best = it_best->BeginCov(2);
    float w,x,y;
    float vertx = 0.0, verty = 0.0, vertweight = 0.0;
    ntrincluster = 0;
    for (int itr=1; itr <= tracksonwire; itr++){ // collect cluster
      it = trackonwire[wnr][itr];
      x = it->BeginX();
      y = it->BeginY();
      w = it->PrimaryWeight();
      float dist2 = (x-x_best)*(x-x_best) + (y-y_best)*(y-y_best);
      float cut2 = wtsig*wtsig*
	(it->BeginCov(0) + it->BeginCov(2) + c0_best + c2_best);
      if(dist2 < cut2){
	vertx += w*x;
	verty += w*y;
	vertweight += w;
	trackincluster[++ntrincluster] = it;
      }
    }

    if (ntrincluster < mintracksinvertex)  break;


    // Make a new GroverVertex in the wire plane and 
    // fill with the candidate tracks
    // Convenient notation but local coordinates ASSUMED!
    
    vertx /= vertweight;  verty /= vertweight;
    GroverVertex vertex(vertx, verty, wire_best->Z());
    vertex.SetWire(wire_best->Number());
    for (int i=1; i<=ntrincluster; i++) vertex.Add(trackincluster[i]);
    
    vertex.EstimateCov(1.0, 0.01);
    
    // Calculate distances from tracks to vertex in xy plane
    vertex.UpdateDistances();
        
    // Sort particle list according to distance from 
    // mean vertex position
    //sort(vertex.particles.begin(), vertex.particles.end(), GroverDistSort);
    
    // FIT VERTEX and see if it's still good
    int vertex_ok = 0;
    int refsize = GeoFit(&vertex);
    
    
    if (refsize >= mintracksinvertex)	{
      // We found a vertex!!!
      vertex_ok = 1;
      nver++;
      vertex.SetIndex(nver); // give it a number
      allVertices->push_back(vertex); // add it to the global vertex list
    }
    
    
    if (vertex_ok == 1) {
      // Flag the used tracks as 'used' in the global particle list
      // by labelling them with the vertex number
      for (GroverParticleList::iterator ipar = vertex.particles.begin();
	   ipar != vertex.particles.end(); ++ipar) {
	if (ipar->UsedInPrimary() != 0) { // not discarded from vertex
	  int id = ipar->Index();
	  (*allParticles)[id-1].SetUsedInPrimary(nver);
	  if (id>=0&&id<1500) grover_rseg_to_rver[id] = nver;
	  //ipar->PropagateBeginTo(vertex.Z()+0.5);
	  if (roverdraw) draw_track(ipar,Black); //max//
	}
      }
      if (roverdraw) draw_vertex(&vertex,Red);
    }
    else {
      // If it's a bad vertex, also flag the _unused_ tracks as 'used'.
      for (GroverParticleList::iterator ipar = vertex.particles.begin();
	   ipar != vertex.particles.end(); ++ipar) {
	if (ipar->UsedInPrimary() != 0)
	  (*allParticles)[(ipar->Index())-1].SetUsedInPrimary(-1);
      }
      if (!vertex.trashcan.empty()) {
	for (GroverParticleList::iterator ipar = vertex.trashcan.begin();
	     ipar != vertex.trashcan.end(); ++ipar) {
	  (*allParticles)[(ipar->Index())-1].SetUsedInPrimary(-1);
	}
      }
    }
    
    
    if (vertex_ok == 1) {
      // Update target position using tracks and vertices
      // -> get current position of wire
      float xt    = wire_best->X();
      float yt    = wire_best->Y();
      float zt    = wire_best->Z();
      float dxt   = wire_best->SizeX();
      float dyt   = wire_best->SizeY();
      float dzt   = wire_best->SizeZ();
      int   nt_tr = wire_best->NParticles();
      int   nt_vx = wire_best->NPrimaries();
      int   nv    = vertex.Size();
      float xv    = vertex.X();
      float yv    = vertex.Y();
      float zv    = vertex.Z();
      // -> calculate new position
      float dx = 0.1, dy = 0.1;
      if ((fabs(xv-xt)<dx)&&(fabs(yv-yt)<dy)){
	xt = xt*(nt_tr+nt_vx)+xv*3*nv;
	yt = yt*(nt_tr+nt_vx)+yv*3*nv;
	zt = zt*nt_vx+zv*3*nv;
	nt_vx += 3*nv;
	xt /= (nt_tr+nt_vx);  
	yt /= (nt_tr+nt_vx);
	zt /=  nt_vx;

	dxt = (dxt*(nt_tr+nt_vx-3*nv) + nv*fabs(xv-xt)) / (nt_tr + nt_vx);
	dyt = (dyt*(nt_tr+nt_vx-3*nv) + nv*fabs(yv-yt)) / (nt_tr + nt_vx);
	dzt = (dzt*(nt_vx-3*nv) + 3*nv*fabs(zv-zt)) / nt_vx;
      }
      // -> set new position
      wire_best->SetX(xt);
      wire_best->SetY(yt);
      wire_best->SetZ(zt);
      wire_best->SetSizeX(dxt);
      wire_best->SetSizeY(dyt);
      wire_best->SetSizeZ(dzt);
      // -> Update global wire position monitoring
      int thiswire = wire_best->Number();
      wire_xpos[thiswire]  = xt;
      wire_ypos[thiswire]  = yt;
      wire_zpos[thiswire]  = zt;
      wire_xsize[thiswire] = dxt;
      wire_ysize[thiswire] = dyt;
      wire_zsize[thiswire] = dzt;
      // after 10k tracks or vertices -> loosen up, man! The target
      // moves slowly, so don't force it to stay in the same place for 
      // ever.
      if (nt_tr+nt_vx > 10000)  { nt_tr=100;  nt_vx = 100;} 
      wire_best->SetNParticles(nt_tr);
      wire_best->SetNPrimaries(nt_vx);
    }
  }



  if (verbose) {
    cout << "*** GroverPrimaries: found "; 
    if (nver == 1) cout << "1 primary vertex" << endl;
    else cout << nver << " primary vertices" << endl;
    cout << "*** GroverPrimaries: finished" << endl; 
  }



  return 0;
}



////////////////////////////////////////////////////////////////////


int GeoFit(GroverVertex* vertex)
{
  // Fit primary vertex using purely geometric method of 
  // Rover 1.05 (I.Kisel/D.Emeliyanov)
  cout.setf(ios::fixed);
  cout.setf(ios::showpoint);
  cout.precision(6);
  typedef struct{
    float cov[4][4];
    float Tcov[2][2];
    float Tpar[2];
    float dzeta[4];
    float zf;
    float ChiInc;
    int flag;
  } TRACK_TYPE;
  TRACK_TYPE atr[ROVERMAXTRACKS];
  int n_track;
  const float dchi2  = 20.0;
  const float Pfalse =  0.1;
  const int   MaxIter=  3;

  float Vcov[3][3];
  float Rk[5];
  float Vpar[3];
  float dG[5][5],KH[5][5],Gk[5][5];
  float Hk[4][5];
  float Kk[5][4];
  double L[4][4];
  float Fk[5][4];
  float lambda,ax,ay,beta1,dz,chi2;
  float res[4],ext[4];
  int i,j,itr,k,nValidTrack; // WOW. Fortran or what! ;-)

  (vertex->trashcan).clear(); // clear trashcan of old particles

  Vpar[0]=vertex->X(); Vpar[1]=vertex->Y(); Vpar[2]=vertex->Z();

  // Fill cov matrices and track parameters
  n_track = 0;
  for (GroverParticleList::iterator ipart = (vertex->particles).begin();
       ipart != (vertex->particles).end(); ++ipart) {
    ipart->PropagateBeginTo(Vpar[2]);
    atr[n_track].cov[0][0] = ipart->BeginCov(0);
    atr[n_track].cov[1][1] = ipart->BeginCov(2);
    atr[n_track].cov[2][2] = ipart->BeginCov(5);
    atr[n_track].cov[3][3] = ipart->BeginCov(9);
    atr[n_track].cov[1][0] = atr[n_track].cov[0][1] = ipart->BeginCov(1);
    atr[n_track].cov[2][0] = atr[n_track].cov[0][2] = ipart->BeginCov(3);
    atr[n_track].cov[2][1] = atr[n_track].cov[1][2] = ipart->BeginCov(4);
    atr[n_track].cov[3][0] = atr[n_track].cov[0][3] = ipart->BeginCov(6);
    atr[n_track].cov[3][1] = atr[n_track].cov[1][3] = ipart->BeginCov(7);
    atr[n_track].cov[3][2] = atr[n_track].cov[2][3] = ipart->BeginCov(8);
    atr[n_track].dzeta[0]  = ipart->BeginX();
    atr[n_track].dzeta[1]  = ipart->BeginY();
    atr[n_track].dzeta[2]  = ipart->BeginSlopeX();
    atr[n_track].dzeta[3]  = ipart->BeginSlopeY();
    atr[n_track].zf        = Vpar[2];
    atr[n_track].Tpar[0]   = atr[n_track].dzeta[2];
    atr[n_track].Tpar[1]   = atr[n_track].dzeta[3];
    for(i=0;i<2;i++) for(j=0;j<2;j++)  
      atr[n_track].Tcov[i][j] = atr[n_track].cov[i+2][j+2];
    atr[n_track].flag=1;
    n_track++;
  }

  // Iteratively fit vertex - number of iterations fixed at 3
  for(int iteration=1; iteration<=MaxIter; iteration++) { 
    // (re)initialize the vertex covariance
    for(i=0;i<3;i++) for(j=0;j<3;j++) Vcov[i][j]=0.0;
    Vcov[0][0] = vertex->Cov(0);  
    Vcov[1][1] = vertex->Cov(2);  
    Vcov[2][2] = vertex->Cov(5);
    
    // Part one: vertex position estimation 
    for(itr=0; itr<n_track; itr++){
      // Construct covariance from vertex and track parts 
      for(i=0;i<5;i++)  for(j=0;j<5;j++)  Gk[i][j] = 0.0;
      for(i=0;i<3;i++)  for(j=0;j<3;j++)  Gk[i][j] = Vcov[i][j];
      for(i=3;i<5;i++)  for(j=3;j<5;j++)  Gk[i][j] = atr[itr].Tcov[i-3][j-3];
      // Construct state vector from vertex and track parts
      for (i=0;i<3;i++) Rk[i]=Vpar[i];
      Rk[3] = atr[itr].Tpar[0];
      Rk[4] = atr[itr].Tpar[1];
      // Linearized measurement matrix 
      dz = atr[itr].zf - Rk[2];
      ax = Rk[3];
      ay = Rk[4];
      for(i=0;i<4;i++)  for(j=0;j<5;j++)  Hk[i][j]=0.0;
      Hk[0][0] = Hk[1][1] = Hk[2][3] = Hk[3][4] = 1.0;
      Hk[0][2] = -ax; Hk[1][2] = -ay;
      Hk[0][3] = Hk[1][4] = dz;
      // Check if track is in the validation region
      for(i=0;i<4;i++)  for(j=0;j<4;j++)  L[i][j]=0.0;
      L[0][0] = Gk[0][0] + ax*ax*Gk[2][2] + dz*dz*Gk[3][3] +
	2.0*(dz*Gk[0][3] - ax*Gk[0][2]    - ax*dz*Gk[2][3]);
      L[1][1] = Gk[1][1] + ay*ay*Gk[2][2] + dz*dz*Gk[4][4] +
	2.0*(dz*Gk[1][4] - ay*Gk[1][2]    - ay*dz*Gk[2][4]);
      L[0][1] = Gk[0][1] + ax*ay*Gk[2][2] + dz*dz*Gk[3][4] +
	dz*(Gk[0][4]+Gk[1][3]) - ay*(Gk[0][2]+dz*Gk[2][3]) -
	ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0] = L[0][1];
      L[2][2] = Gk[3][3];
      L[0][2] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
      L[2][0] = L[0][2];
      L[0][3] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
      L[3][0] = L[0][3];
      L[1][2] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
      L[2][1] = L[1][2];
      L[1][3] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
      L[3][1] = L[1][3];
      L[3][3] = Gk[4][4];
      L[2][3] = L[3][2] = Gk[3][4];
      for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];
      four_dim_inv(L); //  <= Invert L
      ext[0] = Rk[0] + dz*Rk[3];
      ext[1] = Rk[1] + dz*Rk[4];
      ext[2] = Rk[3];
      ext[3] = Rk[4];
      for(i=0;i<4;i++) res[i] = atr[itr].dzeta[i] - ext[i];
      lambda=0.0;
      for(i=0;i<4;i++)  for(j=0;j<4;j++)  lambda += res[i]*res[j]*L[i][j];
      if (lambda>dchi2) {
	atr[itr].flag=0;
	continue;
      }
      else if (atr[itr].flag){
	// Probability of correct classification
	beta1=1.0-Pfalse/(Pfalse+exp(-lambda));
	// Kalman gain
	Fk[0][0] = Gk[0][0] - ax*Gk[0][2] + dz*Gk[0][3];
	Fk[0][1] = Gk[0][1] - ay*Gk[0][2] + dz*Gk[0][4];
	Fk[0][2] = Gk[0][3];
	Fk[0][3] = Gk[0][4];
	Fk[1][0] = Gk[0][1] - ax*Gk[1][2] + dz*Gk[1][3];
	Fk[1][1] = Gk[1][1] - ay*Gk[1][2] + dz*Gk[1][4];
	Fk[1][2] = Gk[1][3];
	Fk[1][3] = Gk[1][4];
	Fk[2][0] = Gk[0][2] - ax*Gk[2][2] + dz*Gk[2][3];
	Fk[2][1] = Gk[1][2] - ay*Gk[2][2] + dz*Gk[2][4];
	Fk[2][2] = Gk[2][3];
	Fk[2][3] = Gk[2][4];
	Fk[3][0] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
	Fk[3][1] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
	Fk[3][2] = Gk[3][3];
	Fk[3][3] = Gk[3][4];
	Fk[4][0] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
	Fk[4][1] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
	Fk[4][2] = Gk[3][4];
	Fk[4][3] = Gk[4][4];
	for(i=0;i<5;i++) for(j=0;j<4;j++){
	  Kk[i][j]=0.0;
	  for(k=0;k<4;k++)  Kk[i][j] += Fk[i][k]*L[k][j];
	}
	// New estimation of the vertex position 
	for(i=0;i<5;i++)  for(j=0;j<4;j++)  Rk[i] += beta1*Kk[i][j]*res[j];
	// New covariance matrix
	for(i=0;i<5;i++) for(j=0;j<5;j++){
	  KH[i][j] = 0.0;
	  for(k=0;k<4;k++) KH[i][j] += Kk[i][k]*Hk[k][j];
	}
	for(i=0;i<5;i++)  for(j=0;j<5;j++){
	  dG[i][j]=0.0;
	  for(k=0;k<5;k++) dG[i][j] += KH[i][k]*Gk[k][j];
	}
	for(i=0;i<5;i++)  for(j=0;j<5;j++)  Gk[i][j] -= beta1*dG[i][j];
	// Extract vertex part
	for(i=0;i<3;i++) Vpar[i] = Rk[i];
	for(i=0;i<3;i++)  for (j=0;j<3;j++)  Vcov[i][j] = Gk[i][j];
      }
    }
    
    // Part two: smoothing track parameters using estimated vertex position
    for(itr=0;itr<n_track;itr++) if(atr[itr].flag){
      // Construct covariance from vertex and track parts
      for(i=0;i<5;i++)  for(j=0;j<5;j++)  Gk[i][j] = 0.0;
      for(i=0;i<3;i++)  for(j=0;j<3;j++)  Gk[i][j] = Vcov[i][j];
      for(i=3;i<5;i++)  for(j=3;j<5;j++)  Gk[i][j] = atr[itr].Tcov[i-3][j-3];
      // Construct state vector from vertex and track parts
      for (i=0;i<3;i++) Rk[i]=Vpar[i];
      Rk[3] = atr[itr].Tpar[0];
      Rk[4] = atr[itr].Tpar[1];
      // Linearized measurement matrix
      dz = atr[itr].zf - Rk[2];
      ax = Rk[3];
      ay = Rk[4];
      for(i=0;i<4;i++)  for(j=0;j<5;j++)  Hk[i][j]=0.0;
      Hk[0][0] = Hk[1][1] = Hk[2][3] = Hk[3][4] = 1.0;
      Hk[0][2] = -ax;
      Hk[1][2] = -ay;
      Hk[0][3] = Hk[1][4] = dz;
      // Check if track is still in the validation region
      for(i=0;i<4;i++)  for(j=0;j<4;j++)  L[i][j] = 0.0;
      L[0][0] = Gk[0][0] + ax*ax*Gk[2][2] + dz*dz*Gk[3][3] +
	2.0*(dz*Gk[0][3] - ax*Gk[0][2]    - ax*dz*Gk[2][3]);
      L[1][1] = Gk[1][1] + ay*ay*Gk[2][2] + dz*dz*Gk[4][4] +
	2.0*(dz*Gk[1][4] - ay*Gk[1][2]    - ay*dz*Gk[2][4]);
      L[0][1] = Gk[0][1] + ax*ay*Gk[2][2] + dz*dz*Gk[3][4] + 
	dz*(Gk[0][4]+Gk[1][3]) - ay*(Gk[0][2]+dz*Gk[2][3]) - 
	ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0] = L[0][1];
      L[2][2] = Gk[3][3];
      L[0][2] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
      L[2][0] = L[0][2];
      L[0][3] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
      L[3][0] = L[0][3];
      L[1][2] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
      L[2][1] = L[1][2];
      L[1][3] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
      L[3][1] = L[1][3];
      L[3][3] = Gk[4][4];
      L[2][3] = L[3][2] = Gk[3][4];
      for(i=0;i<4;i++)  for (j=0;j<4;j++)  L[i][j]+=atr[itr].cov[i][j];
      four_dim_inv(L); // <= Invert L
      ext[0] = Rk[0] + dz*Rk[3];
      ext[1] = Rk[1] + dz*Rk[4];
      ext[2] = Rk[3];
      ext[3] = Rk[4];
      for(i=0;i<4;i++) res[i] = atr[itr].dzeta[i] - ext[i];
      lambda=0.0;
      for(i=0;i<4;i++)  for(j=0;j<4;j++)  lambda += res[i]*res[j]*L[i][j];
      if(lambda>dchi2) {
	// discard track
	atr[itr].flag=0;
	continue;
      }
      else{
	if(iteration==MaxIter) atr[itr].ChiInc=lambda;
	// Probability of correct classification
	beta1=1.0-Pfalse/(Pfalse+exp(-lambda));
	// Kalman gain
	Fk[0][0] = Gk[0][0] - ax*Gk[0][2] + dz*Gk[0][3];
	Fk[0][1] = Gk[0][1] - ay*Gk[0][2] + dz*Gk[0][4];
	Fk[0][2] = Gk[0][3];
	Fk[0][3] = Gk[0][4];
	Fk[1][0] = Gk[0][1] - ax*Gk[1][2] + dz*Gk[1][3];
	Fk[1][1] = Gk[1][1] - ay*Gk[1][2] + dz*Gk[1][4];
	Fk[1][2] = Gk[1][3];
	Fk[1][3] = Gk[1][4];
	Fk[2][0] = Gk[0][2] - ax*Gk[2][2] + dz*Gk[2][3];
	Fk[2][1] = Gk[1][2] - ay*Gk[2][2] + dz*Gk[2][4];
	Fk[2][2] = Gk[2][3];
	Fk[2][3] = Gk[2][4];
	Fk[3][0] = Gk[0][3] - ax*Gk[2][3] + dz*Gk[3][3];
	Fk[3][1] = Gk[1][3] - ay*Gk[2][3] + dz*Gk[3][4];
	Fk[3][2] = Gk[3][3];
	Fk[3][3] = Gk[3][4];
	Fk[4][0] = Gk[0][4] - ax*Gk[2][4] + dz*Gk[3][4];
	Fk[4][1] = Gk[1][4] - ay*Gk[2][4] + dz*Gk[4][4];
	Fk[4][2] = Gk[3][4];
	Fk[4][3] = Gk[4][4];
	for(i=0;i<5;i++)  for(j=0;j<4;j++){
	  Kk[i][j] = 0.0;
	  for(k=0;k<4;k++) Kk[i][j] += Fk[i][k]*L[k][j];
	}
	// New estimation	
	for(i=0;i<5;i++)  for(j=0;j<4;j++)  Rk[i] += beta1*Kk[i][j]*res[j];
	// New covariance matrix
	for(i=0;i<5;i++)  for(j=0;j<5;j++) {
	  KH[i][j] = 0.0;
	  for(k=0;k<4;k++) KH[i][j] += Kk[i][k]*Hk[k][j];
	}
	for(i=0;i<5;i++)  for(j=0;j<5;j++) {
	  dG[i][j] = 0.0;
	  for(k=0;k<5;k++) dG[i][j] += KH[i][k]*Gk[k][j];
	}
	for(i=0;i<5;i++)  for(j=0;j<5;j++)  Gk[i][j] -= beta1*dG[i][j];
	// Extract track and vertex parts
	atr[itr].Tpar[0] = Rk[3];
	atr[itr].Tpar[1] = Rk[4];
      }
    }
  } // finished iterations

  chi2 = 0.0;
  nValidTrack = 0;
  for(int itr=0; itr<n_track; itr++) 
    if (atr[itr].flag != 0) {
      chi2 += atr[itr].ChiInc;
      nValidTrack++;
    }
  chi2 /= (4.0*nValidTrack - 3.0 - 2.0*nValidTrack);

  // Write vertex parameters and update lists
  vertex->SetChi2(chi2);
  vertex->SetX(Vpar[0]);
  vertex->SetY(Vpar[1]);
  vertex->SetZ(Vpar[2]);
  vertex->SetCov(0, Vcov[0][0]);     
  vertex->SetCov(1, Vcov[0][1]);
  vertex->SetCov(2, Vcov[1][1]);
  vertex->SetCov(3, Vcov[0][2]);     
  vertex->SetCov(4, Vcov[1][2]);
  vertex->SetCov(5, Vcov[2][2]);
  int refntrincluster = 0; // count the reference (good quality) tracks
  itr = 0;
  for (GroverParticleList::iterator ipart = (vertex->particles).begin();
       ipart != (vertex->particles).end(); ++ipart) {
    //itr = ipart->Index();
    // if it's been excluded from vertex, mark as unused and move to trash
    if (atr[itr].flag == 0) {
      ipart->SetUsedInPrimary(0);
      (vertex->trashcan).push_back(*ipart);
      (vertex->particles).erase(ipart); ipart--;
    }
    else {
      ipart->SetUsedInPrimary(1);
      if (ipart->Nhit() >= 5) refntrincluster++; // 7
    }
    itr++;
  }

  // return value is number of refset tracks
  return refntrincluster; 

}


/////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////


void GroverSecondaryFit (GroverVertex* vertex)
{
  // Forces the tracks in a two-track vertex to be a 
  // secondary, estimates position and covariances.

  typedef struct{
    float cov[4][4];
    float Tcov[2][2];
    float Tpar[2];
    float dzeta[4];
    float zf;
  } TRACK_TYPE, *pTRACK_TYPE;
  
  TRACK_TYPE atr[2]; 
  float Vcov[3][3];
  float Rk[5];
  float Vpar[3];
  float dG[5][5],KH[5][5],Gk[5][5];
  float Hk[4][5];
  float Kk[5][4];
  double L[4][4];
  float Fk[5][4];
  float ax,ay,dz,detr,zref=5.0;
  float res[4],ext[4];
  int i,j,itr,k,jjj;

  //  GroverParticleList::iterator trackincluster[2];
  GroverParticleList::iterator it1, it2;

  if (vertex->Size() >= 2) {
    it1 = &(vertex->particles)[0];
    it2 = &(vertex->particles)[1];
  }
  else return;
  //trackincluster[2] = 
  //it2 = ++it1;
   
  vertex->SetMass(0.0); vertex->SetMassError(0.0);
  it1->PropagateBeginTo(zref);
  it2->PropagateBeginTo(zref);
  
  // estimate position of vertex
  detr = (it2->BeginSlopeX() - it1->BeginSlopeX())*
         (it2->BeginSlopeX() - it1->BeginSlopeX()) 
       + (it2->BeginSlopeY() - it1->BeginSlopeY())*
         (it2->BeginSlopeY() - it1->BeginSlopeY());
  vertex->SetZ(  (it2->BeginSlopeX() - it1->BeginSlopeX())*
                 (it2->BeginX()      - it1->BeginX())
               + (it2->BeginSlopeY() - it1->BeginSlopeY())*
                 (it2->BeginY()      - it1->BeginY())    );
  vertex->SetZ(zref - vertex->Z()/detr);
  dz = vertex->Z() - zref;
  vertex->SetX(0.5*(  it2->BeginX() + dz*(it2->BeginSlopeX())
		    + it1->BeginX() + dz*(it1->BeginSlopeX())) );
  vertex->SetY(0.5*(  it2->BeginY() + dz*(it2->BeginSlopeY())
		    + it1->BeginY() + dz*(it1->BeginSlopeY())) );


  // Fill parameters for fit
  Vpar[0] = vertex->X();   Vpar[1] = vertex->Y();   Vpar[2] = vertex->Z();

  //float newcov[21] = {3,0,3,0,0,25,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  float newcov[21] = {1,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
  
  vertex->SetCov(newcov);

  atr[0].cov[0][0] = it1->BeginCov(0);
  atr[0].cov[1][1] = it1->BeginCov(2);
  atr[0].cov[2][2] = it1->BeginCov(5);
  atr[0].cov[3][3] = it1->BeginCov(9);
  atr[0].cov[1][0] = atr[0].cov[0][1] = it1->BeginCov(1);
  atr[0].cov[2][0] = atr[0].cov[0][2] = it1->BeginCov(3);
  atr[0].cov[2][1] = atr[0].cov[1][2] = it1->BeginCov(4);
  atr[0].cov[3][0] = atr[0].cov[0][3] = it1->BeginCov(6);
  atr[0].cov[3][1] = atr[0].cov[1][3] = it1->BeginCov(7);
  atr[0].cov[3][2] = atr[0].cov[2][3] = it1->BeginCov(8);
  atr[0].dzeta[0]  = it1->BeginX();
  atr[0].dzeta[1]  = it1->BeginY();
  atr[0].dzeta[2]  = it1->BeginSlopeX();
  atr[0].dzeta[3]  = it1->BeginSlopeY();
  atr[0].zf        = zref;
  atr[0].Tpar[0]   = atr[0].dzeta[2];
  atr[0].Tpar[1]   = atr[0].dzeta[3];
  for(i=0;i<2;i++) for(j=0;j<2;j++) atr[0].Tcov[i][j] = atr[0].cov[i+2][j+2];  
  atr[1].cov[0][0] = it2->BeginCov(0);
  atr[1].cov[1][1] = it2->BeginCov(2);
  atr[1].cov[2][2] = it2->BeginCov(5);
  atr[1].cov[3][3] = it2->BeginCov(9);
  atr[1].cov[1][0] = atr[1].cov[0][1] = it2->BeginCov(1);
  atr[1].cov[2][0] = atr[1].cov[0][2] = it2->BeginCov(3);
  atr[1].cov[2][1] = atr[1].cov[1][2] = it2->BeginCov(4);
  atr[1].cov[3][0] = atr[1].cov[0][3] = it2->BeginCov(6);
  atr[1].cov[3][1] = atr[1].cov[1][3] = it2->BeginCov(7);
  atr[1].cov[3][2] = atr[1].cov[2][3] = it2->BeginCov(8);
  atr[1].dzeta[0]  = it2->BeginX();
  atr[1].dzeta[1]  = it2->BeginY();
  atr[1].dzeta[2]  = it2->BeginSlopeX();
  atr[1].dzeta[3]  = it2->BeginSlopeY();
  atr[1].zf        = zref;
  atr[1].Tpar[0]   = atr[1].dzeta[2];
  atr[1].Tpar[1]   = atr[1].dzeta[3];
  for(i=0;i<2;i++) for(j=0;j<2;j++)  atr[1].Tcov[i][j] = atr[1].cov[i+2][j+2];
  
  for(jjj=0;jjj<1;jjj++){
    for(i=0;i<3;i++) for(j=0;j<3;j++) Vcov[i][j]=0.0;
    Vcov[0][0]=vertex->Cov(0); 
    Vcov[1][1]=vertex->Cov(2);
    Vcov[2][2]=vertex->Cov(5);

    // loop over track pair
    for(itr=0;itr<2;itr++)
      {
	// Construct covariance from vertex and track parts
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]=0.0;
	for(i=0;i<3;i++) for(j=0;j<3;j++) Gk[i][j]=Vcov[i][j];
	for(i=3;i<5;i++) for(j=3;j<5;j++) Gk[i][j]=atr[itr].Tcov[i-3][j-3];
	// Construct state vector from vertex and track parts
	for (i=0;i<3;i++) Rk[i]=Vpar[i];
	Rk[3]=atr[itr].Tpar[0]; 
	Rk[4]=atr[itr].Tpar[1];
	// Linearized measurement matrix 
	dz=atr[itr].zf-Rk[2];ax=Rk[3];ay=Rk[4];      
	for(i=0;i<4;i++) for(j=0;j<5;j++) Hk[i][j]=0.0;
	Hk[0][0]=Hk[1][1]=Hk[2][3]=Hk[3][4]=1.0;
	Hk[0][2]=-ax;Hk[1][2]=-ay;Hk[0][3]=dz;Hk[1][4]=dz;
	// Check if track is in the validation region
	for(i=0;i<4;i++) for(j=0;j<4;j++) L[i][j]=0.0;
	L[0][0]=Gk[0][0]+ax*ax*Gk[2][2]+dz*dz*Gk[3][3]+
	  2.0*(dz*Gk[0][3]-ax*Gk[0][2]-ax*dz*Gk[2][3]);
	L[1][1]=Gk[1][1]+ay*ay*Gk[2][2]+dz*dz*Gk[4][4]+
	  2.0*(dz*Gk[1][4]-ay*Gk[1][2]-ay*dz*Gk[2][4]);
	L[0][1]=Gk[0][1]+ax*ay*Gk[2][2]+dz*dz*Gk[3][4]+dz*(Gk[0][4]+Gk[1][3])-
	  ay*(Gk[0][2]+dz*Gk[2][3])-ax*(Gk[1][2]+dz*Gk[2][4]);
	L[1][0]=L[0][1];L[2][2]=Gk[3][3];
	L[0][2]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	L[2][0]=L[0][2];L[0][3]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	L[3][0]=L[0][3];L[1][2]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	L[2][1]=L[1][2];L[1][3]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	L[3][1]=L[1][3];L[3][3]=Gk[4][4];L[2][3]=L[3][2]=Gk[3][4];
	for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];
	four_dim_inv(L); // <= invert L

	ext[0]=Rk[0]+dz*Rk[3];ext[1]=Rk[1]+dz*Rk[4];ext[2]=Rk[3];ext[3]=Rk[4];
	for(i=0;i<4;i++) res[i]=atr[itr].dzeta[i]-ext[i];
	
	// kalman gain
	Fk[0][0]=Gk[0][0]-ax*Gk[0][2]+dz*Gk[0][3];
	Fk[0][1]=Gk[0][1]-ay*Gk[0][2]+dz*Gk[0][4];
	Fk[0][2]=Gk[0][3];Fk[0][3]=Gk[0][4];
	Fk[1][0]=Gk[0][1]-ax*Gk[1][2]+dz*Gk[1][3];
	Fk[1][1]=Gk[1][1]-ay*Gk[1][2]+dz*Gk[1][4];
	Fk[1][2]=Gk[1][3];Fk[1][3]=Gk[1][4];
	Fk[2][0]=Gk[0][2]-ax*Gk[2][2]+dz*Gk[2][3];
	Fk[2][1]=Gk[1][2]-ay*Gk[2][2]+dz*Gk[2][4];
	Fk[2][2]=Gk[2][3];Fk[2][3]=Gk[2][4];
	Fk[3][0]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
	Fk[3][1]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
	Fk[3][2]=Gk[3][3];Fk[3][3]=Gk[3][4];
	Fk[4][0]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
	Fk[4][1]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
	Fk[4][2]=Gk[3][4];Fk[4][3]=Gk[4][4];
	for(i=0;i<5;i++) for(j=0;j<4;j++)
	  {
	    Kk[i][j]=0.0;
	    for(k=0;k<4;k++) Kk[i][j]+=Fk[i][k]*L[k][j];
	  }
	
	// New estimate of vertex position
	for(i=0;i<5;i++) for(j=0;j<4;j++) Rk[i]+=Kk[i][j]*res[j];
	
	// new cov matrix
	for(i=0;i<5;i++) for(j=0;j<5;j++)
	  {
	    KH[i][j]=0.0;
	    for(k=0;k<4;k++) KH[i][j]+=Kk[i][k]*Hk[k][j];
	  }
	for(i=0;i<5;i++) for(j=0;j<5;j++)
	  {
	    dG[i][j]=0.0;
	    for(k=0;k<5;k++) dG[i][j]+=KH[i][k]*Gk[k][j];
	  }
	for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]-=dG[i][j];
	// Extract vertex part
	for(i=0;i<3;i++) Vpar[i]=Rk[i];
	for(i=0;i<3;i++) for (j=0;j<3;j++) Vcov[i][j]=Gk[i][j];
      }

    // Smoothing
    for(itr=0;itr<2;itr++) {
      // Construct covariance from vertex and track parts
      for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]=0.0;
      for(i=0;i<3;i++) for(j=0;j<3;j++) Gk[i][j]=Vcov[i][j];
      for(i=3;i<5;i++) for(j=3;j<5;j++) Gk[i][j]=atr[itr].Tcov[i-3][j-3];
      // Linearized measurement matrix
      for (i=0;i<3;i++) Rk[i]=Vpar[i];
      Rk[3]=atr[itr].Tpar[0];Rk[4]=atr[itr].Tpar[1];
      
      dz = atr[itr].zf-Rk[2];
      ax = Rk[3];
      ay = Rk[4];
      for(i=0;i<4;i++) for(j=0;j<5;j++) Hk[i][j]=0.0;
      Hk[0][0]=Hk[1][1]=Hk[2][3]=Hk[3][4]=1.0;
      Hk[0][2]=-ax;Hk[1][2]=-ay;Hk[0][3]=dz;Hk[1][4]=dz;
      for(i=0;i<4;i++) for(j=0;j<4;j++) L[i][j]=0.0;
      L[0][0]=Gk[0][0]+ax*ax*Gk[2][2]+dz*dz*Gk[3][3]+
	2.0*(dz*Gk[0][3]-ax*Gk[0][2]-ax*dz*Gk[2][3]);
      L[1][1]=Gk[1][1]+ay*ay*Gk[2][2]+dz*dz*Gk[4][4]+
	2.0*(dz*Gk[1][4]-ay*Gk[1][2]-ay*dz*Gk[2][4]);
      L[0][1]=Gk[0][1]+ax*ay*Gk[2][2]+dz*dz*Gk[3][4]+dz*(Gk[0][4]+Gk[1][3])-
	ay*(Gk[0][2]+dz*Gk[2][3])-ax*(Gk[1][2]+dz*Gk[2][4]);
      L[1][0]=L[0][1];L[2][2]=Gk[3][3];
      L[0][2]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
      L[2][0]=L[0][2];L[0][3]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
      L[3][0]=L[0][3];L[1][2]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
      L[2][1]=L[1][2];L[1][3]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
      L[3][1]=L[1][3];L[3][3]=Gk[4][4];L[2][3]=L[3][2]=Gk[3][4];
      for(i=0;i<4;i++) for (j=0;j<4;j++) L[i][j]+=atr[itr].cov[i][j];      
      four_dim_inv(L); // <= invert L
      ext[0]=Rk[0]+dz*Rk[3];ext[1]=Rk[1]+dz*Rk[4];ext[2]=Rk[3];ext[3]=Rk[4];
      for(i=0;i<4;i++) res[i]=atr[itr].dzeta[i]-ext[i];
      
      // kalman gain
      Fk[0][0]=Gk[0][0]-ax*Gk[0][2]+dz*Gk[0][3];
      Fk[0][1]=Gk[0][1]-ay*Gk[0][2]+dz*Gk[0][4];
      Fk[0][2]=Gk[0][3];Fk[0][3]=Gk[0][4];
      Fk[1][0]=Gk[0][1]-ax*Gk[1][2]+dz*Gk[1][3];
      Fk[1][1]=Gk[1][1]-ay*Gk[1][2]+dz*Gk[1][4];
      Fk[1][2]=Gk[1][3];Fk[1][3]=Gk[1][4];
      Fk[2][0]=Gk[0][2]-ax*Gk[2][2]+dz*Gk[2][3];
      Fk[2][1]=Gk[1][2]-ay*Gk[2][2]+dz*Gk[2][4];
      Fk[2][2]=Gk[2][3];Fk[2][3]=Gk[2][4];
      Fk[3][0]=Gk[0][3]-ax*Gk[2][3]+dz*Gk[3][3];
      Fk[3][1]=Gk[1][3]-ay*Gk[2][3]+dz*Gk[3][4];
      Fk[3][2]=Gk[3][3];Fk[3][3]=Gk[3][4];
      Fk[4][0]=Gk[0][4]-ax*Gk[2][4]+dz*Gk[3][4];
      Fk[4][1]=Gk[1][4]-ay*Gk[2][4]+dz*Gk[4][4];
      Fk[4][2]=Gk[3][4];Fk[4][3]=Gk[4][4];
      for(i=0;i<5;i++) for(j=0;j<4;j++)
	{
	  Kk[i][j]=0.0;
	  for(k=0;k<4;k++) Kk[i][j]+=Fk[i][k]*L[k][j];
	}
      // new estimate
      for(i=0;i<5;i++) for(j=0;j<4;j++) Rk[i]+=Kk[i][j]*res[j];
      // new covariance matrix
      for(i=0;i<5;i++) for(j=0;j<5;j++)
	{
	  KH[i][j]=0.0;
	  for(k=0;k<4;k++) KH[i][j]+=Kk[i][k]*Hk[k][j];
	}
      
      for(i=0;i<5;i++) for(j=0;j<5;j++)
	{
	  dG[i][j]=0.0;
	  for(k=0;k<5;k++) dG[i][j]+=KH[i][k]*Gk[k][j];
	}
      
      for(i=0;i<5;i++) for(j=0;j<5;j++) Gk[i][j]-=dG[i][j];
      
      // extract track and vtx parts
      atr[itr].Tpar[0]=Rk[3];atr[itr].Tpar[1]=Rk[4];
      Vpar[0]=Rk[0];Vpar[1]=Rk[1];Vpar[2]=Rk[2];
    }
  }
  // Write vertex position and covariances
  vertex->SetX(Vpar[0]);
  vertex->SetY(Vpar[1]);
  vertex->SetZ(Vpar[2]);
  vertex->SetCov(0, Vcov[0][0]); 
  vertex->SetCov(1, Vcov[0][1]);
  vertex->SetCov(2, Vcov[1][1]);
  vertex->SetCov(3, Vcov[0][2]);
  vertex->SetCov(4, Vcov[1][2]);
  vertex->SetCov(5, Vcov[2][2]);

  return;
}


///////////////////////////////////////////////////////////////////////////


