/* Some tools for Setups with fired/non fired cells.
   (See fired_cells.hh).

   Includes:
   1) Write_Data_To_Fired_Setup_Cell<ISC_Or_ICC> write_data;
      // function object for writing data
      // ISC_Or_ICC can be YourSetup::ISectionCell or 
      // YourSetup::IChamberCell
      write_data(data, isc_or_icc); // this places data 
      // to *(isc_or_icc.icell).second;
      // In addition if this fires the cell, the icc of this cell is
      // added to the vector of fired cells for this section and 
      // increments the number of hits in the cell
   Or you may just use
     write_data_to_fired_setup_cell(data, isc_or_icc);
   this directly WRITES the data. (Function returns void). 

   2) Resetting all the cells in the vector of fired cells
   and clearing the vector itself.
   
   To reset the whole setup and prepare it for the new event,
   you do not need to reset every cell in the setup with, e.g.
   reset_cell_data_in(setup, default_value); from setup_tools.hh.
   One needs to reset only those cells which have been fired previously,
   i.e. those which are in the fired_cells vector.
   If the number of fired cells is low (occupancy is low) it can be much faster.

   One may perform this task with
      ... { // loop over events
        // in the beginning of each event:
        reset_fired_cells_in(setup);
	...
      }
   Or one may write reset_fired_cells_in(chamber),
                    reset_fired_cells_in(detector); // if all
		    // Setups in the detector have fired/non fired cells

   If one needs to clear only the fired_cells vector without resetting the
   corresponding cells, one may use
      ... { // loop over events
        // in the beginning of each event:
        reset_fired_cell_vectors_in(setup);
	...
      }
   Or               reset_fired_cell_vectors_in(chamber),
                    reset_fired_cell_vectors_in(detector); // again if all
		    // Setups in the detector have fired/non fired cells

   3)   calculate_cell_occupancy_in(setup, n_events);
      Or you may write
        calculate_cell_occupancy_in(chamber, n_events);
        calculate_cell_occupancy_in(section, n_events);

	calculate_cell_occupancy_in(detector, n_events); // if all
	// Setups in the detector have fired/non fired cells
      One function works in all 4 cases.
      It fills the "occupancy" field in all cells knowing their
      "n_hits" and the given "n_events".

   4) calculate_average_occupancy_and_mark_working_cells_in(setup,
         low, high, bad_occ, noisy_occ);
      (Or you may do it within one chamber, or within the whole detector
      if all setups in it have fired/non fired cells).
      Explanation of parameters:
      low, high (e.g. 0.3 and 0.7): average occupancy 
                   will be calculated among all occupancies in
                   the section except 30% of lowest and 30% of
                   highest values
      bad_occ, noisy_occ (e.g. 0.25 and 4):
                    cells are considered as bad or noisy if their
		    occupancy is less than 0.25*(average value) 
		    or higher than 4*(average value) respectively.

    Requirements:
      write_data_to_fired_setup_cell(data, isc_or_icc); works if:
      1) classes derived from Fired_Cell provide is_fired() method
      with return type convertible to bool. This function should check
      whether to increment the number of hits for the cell and puch_back()
      the cell to the fired_cells vector;
      2) class CellType has "data" field of type CellType::DataType.
   */
   
     
// Author V.Balagura (balagura@mail.desy.de)   (18.02.99).

#ifndef _fired_cell_tools_hh
#define _fired_cell_tools_hh 1

#include "ptset/detector_tools.hh"
#include "ptset/fired_cells.hh"
#include <numeric> // for accumulate in avr. occupancy calculation

/* ----------------- Write_Data_To_Fired_Setup_Cell ----------------- */
template<class ISC_Or_ICC>
class Write_Data_To_Fired_Setup_Cell {
public:
  typedef typename ISC_Or_ICC::ContainerType::Cell Cell;
  void operator()(const typename Cell::DataType& data, 
		  const ISC_Or_ICC& isc) {
    Cell& cell = (*isc.icell).second;
    cell.data = data;
    if ( cell.is_fired() ) {
      ++cell.n_hits;
      (*isc.isection).fired_cells.push_back(isc.icell);
    }
  }
};
template<class DataType, class ISC_Or_ICC>
inline void write_data_to_fired_setup_cell(const DataType& data, 
					   const ISC_Or_ICC& isc) {
  Write_Data_To_Fired_Setup_Cell<ISC_Or_ICC>()(data, isc);
  // DataType should be convertible to Cell::DataType
}
  
/* -------------------- Reset_Fired_Cell_Vectors -------------------- */
struct Reset_Fired_Cell_Vectors {
  template<class AnyISection>
  void operator()(AnyISection is) const { (*is).fired_cells.clear(); }
};
/* --------------- reset_fired_cell_vectors_in --------------- */
template <class T> 
inline void reset_fired_cell_vectors_in(T& x) {
  for_each_section(Reset_Fired_Cell_Vectors(), x);
}

/* -------------------- Reset_Fired_Cells -------------------- */
template <class AnyISection>
struct Reset_Fired_Cells {
  typedef typename iterator_traits<AnyISection>::value_type::Cell::DataType
  Data;
  Data data;
  Reset_Fired_Cells(const Data& default_data) : data(default_data) {}
  void operator()(AnyISection is) const {
    for (unsigned int i=0; i<(*is).fired_cells.size(); ++i) 
      (*(*is).fired_cells[i]).second.data = data;
    (*is).fired_cells.clear();
  }
};
/* --------------- reset_fired_cells_in --------------- */
template <class T, class DataType> 
inline void reset_fired_cells_in(T& x, const DataType& data) {
  for_each_section(Reset_Fired_Cells<typename T::ISection>(data), x);
  // DataType should be convertible to T::Cell::DataType
}

/* -------------------- Calculate_Cell_Occupancy -------------------- */
struct Calculate_Cell_Occupancy {
  unsigned int n_ev;
  Calculate_Cell_Occupancy(unsigned int n_events) : n_ev(n_events) {}

  template<class AnyICell>
  void operator()(AnyICell ic) const {
    typedef typename iterator_traits<AnyICell>::value_type::second_type 
      Cell;
    Cell& cell = (*ic).second;
    cell.occupancy = double(cell.n_hits) / n_ev;
  }
};
/* --------------- calculate_cell_occupancy_in --------------- */
template <class T> 
inline void calculate_cell_occupancy_in(T& x, unsigned int n_events) {
  for_each_cell(Calculate_Cell_Occupancy(n_events), x);
}

/* - Calculate_Average_Occupancy_And_Mark_Working_Cells_In_One_Section - */
struct Calculate_Average_Occupancy_And_Mark_Working_Cells_In_One_Section {
  double low, high; // e.g. 0.3 and 0.7: average occupancy 
                   // will be calculated among all occupancies in
                   // the section except 30% of the lowest and 30% of the
                   // highest values
  double bad_occupancy, noisy_occupancy;//e.g. 0.25 and 4: cells are considered
                                        // as bad or noisy if their occupancy
                                        // is less than 0.25*(average value) 
                                        // or higher than 4*(average value)
                                        // respectively.
  Calculate_Average_Occupancy_And_Mark_Working_Cells_In_One_Section
    (double low_limit, 
     double high_limit,
     double bad_occupancy_limit,
     double noisy_occupancy_limit) : low(low_limit), high(high_limit), 
                                    bad_occupancy(bad_occupancy_limit), 
                                    noisy_occupancy(noisy_occupancy_limit)
  {}

  template <class AnyISection>
  void operator()(AnyISection is) const {
    typedef typename iterator_traits<AnyISection>::value_type Section;
    typedef typename Section::Cell Cell;
    vector<double> occ;
    typename Section::ICell icell;
    Section& section = *is;
    for (icell=section.begin(); icell!=section.end(); ++icell) {
      Cell& cell = (*icell).second;
      if ( cell.is_read_out ) occ.push_back(cell.occupancy); 
    }
    sort(occ.begin(), occ.end());
    int i1, i2;
    i1 = int( occ.size() * low ); 
    i2 = int( occ.size() * high );
    if ( i1 < i2 ) {
      section.average_occupancy = 
	accumulate(occ.begin()+i1, 
		   occ.begin()+i2, 0.) / double(i2-i1);
      double x1, x2;
      x1 = section.average_occupancy * bad_occupancy;
      x2 = section.average_occupancy * noisy_occupancy;
      for (icell=section.begin(); icell!=section.end(); ++icell) {
	Cell& cell = (*icell).second;
	if ( cell.is_read_out ) {
	  if      (cell.occupancy < x1) cell.state = Fired_Cell::BAD;
	  else if (cell.occupancy > x2) cell.state = Fired_Cell::NOISY;
	  else                          cell.state = Fired_Cell::Ok;
	}
      }
    }
  }
};
/*  calculate_average_occupancy_and_mark_working_cells_in */
template <class T> 
inline void calculate_average_occupancy_and_mark_working_cells_in
(T& x, double low_limit, double high_limit,
double bad_occupancy_limit, double noisy_occupancy_limit) {
  Calculate_Average_Occupancy_And_Mark_Working_Cells_In_One_Section
    f(low_limit, high_limit, bad_occupancy_limit, noisy_occupancy_limit);
  for_each_section(f, x);
}


#endif
