/* Classes describing non geometrical information of High-Pt system.
   The resulting Setup classes are:
      HPT_Cell;
      HPT_Section;
      HPT_Chamber;
      HPT_Setup;
      */

// Modification made in the beginning of May, 2000.
// In order to reflect the structure of high-pt chambers which can be
// divided into inner/outer and into 3 superlayers, HPT_Setup now
// contains additionally two functions:
//
// const vector<vector<vector<HPT_Setup::IChamber> > >& ichambers() const;
// const HPT_ChamberID* position(const string& name) const;
//
// The first one returns container of HPT_Setup::IChamber iterators ordered
// in such a way that 
// ichambers()[ilayer][ipart][ichamber_plus_or_minus_x]
// points to the chamber in the layer "ilayer" = 0, 1, 2, in the part "ipart"
// (inner: ipart=0; outer: ipart=1) and -x if ichamber_plus_or_minus_x=0
// or +x if =1.

// The second function performs the opposite operation: given 
// the name of the chamber it returns the pointer to the structure HPT_ChamberID
// which contains ilayer, ipart, ichamber corresponding to the
// ilayer, ipart and ichamber_plus_or_minus_x above. If the chamber with
// the given name is not found, the zero pointer is returned.

// In addition one more function was added to HPT_Setup:
//  void clear();
// It deletes all components in HPT_Setup and makes it empty.

// Since there is one High-Pt setup in HERA-B and this is fixed the following
// global object is defined:
//
//   extern HPT_Setup hpt_setup;
//

// Modification made on 12.09.00:
// the functions 
//   void write_hits(ostream& os) and
//   void read_hits (istream& is) have been added.
// They deal only with hits (e.g. per event); 
// the structure of the setup is not written/read.
// Thus one should ensure that the structures are really the same in 
// writing and in reading.

// Author V.Balagura (balagura@mail.desy.de)   (09.04.98).

#ifndef _high_pt_hh
#define _high_pt_hh 1

#include "ptset/setup.hh"
#include "ptset/dummy.hh"
#include "ptset/fired_cells.hh"
//#include "ptset/global_error.hh"

class HPT_CellT : public Fired_Cell {
public:
  typedef bool DataType;
  DataType data;
  bool is_fired() { return data; }
  HPT_CellT() : data(false) {}
};

struct HPT_SetupT {
  static const string& setup_type() { static string x("high-pt"); return x; }
};  
// typedef Section_With_Fired_Cells<HPT_CellT> HPT_SectionT;
struct HPT_SectionT : public Section_With_Fired_Cells<HPT_CellT> {};

typedef SetupCell<HPT_CellT>                                 HPT_Cell;
typedef SetupSection<HPT_CellT, HPT_SectionT>                HPT_Section;
typedef SetupChamber<HPT_CellT, HPT_SectionT, Dummy>         HPT_Chamber;

struct HPT_ChamberID {
  unsigned int ilayer, ipart, ichamber;
  HPT_ChamberID() {}
  HPT_ChamberID(unsigned int fIlayer, unsigned int fIpart,
		unsigned int fIchamber) : 
    ilayer(fIlayer), ipart(fIpart), ichamber(fIchamber) {}
};  

class HPT_Setup : public Setup<HPT_CellT, HPT_SectionT, Dummy, HPT_SetupT> {
public:
  const vector<vector<vector<HPT_Setup::IChamber> > >& ichambers() const;
  // ichambers()[ilayer][ipart][ichamber_plus_or_minus_x]
  const HPT_ChamberID* position(const string& name) const;
  // position()->ilayer,ipart,ichamber where ichamber (usually) =0 for x-,
  // 1 for x+ (it has the type "unsigned int" and not HPT_Setup::IChamber)
  void read(istream& is);

  // the following function is used in read(); you may do not use it directly
  void create_chambers(const vector<vector<vector<string> > >& names);
  // create chambers with names names[ilayer][ipart][ich] and make links
  // to mIChambers, mID
  void clear();
  // deletes all components from HPT_Setup so that it becomes empty

  void write_hits(ostream& os); // there are no checks that the Setups used for
  void read_hits (istream& is); // writing and then for reading have the same
  // structure
private:
  vector<vector<vector<HPT_Setup::IChamber> > > mIChambers;
  map<string, HPT_ChamberID> mID;
  void cross_reference();
};

extern HPT_Setup hpt_setup;

// ------------------------- inline functions -------------------------
inline bool operator<(const HPT_ChamberID& x, const HPT_ChamberID& y) {
  return
    x.ilayer < y.ilayer ||
    (x.ilayer == y.ilayer &&
     (x.ipart < y.ipart || 
      (x.ipart == y.ipart && 
       (x.ichamber < y.ichamber))));
}
inline bool operator==(const HPT_ChamberID& x, const HPT_ChamberID& y) {
  return x.ilayer == y.ilayer && x.ipart == y.ipart && x.ichamber == y.ichamber;
}

inline const vector<vector<vector<HPT_Setup::IChamber> > >&
HPT_Setup::ichambers() const { return mIChambers; }

inline const HPT_ChamberID* HPT_Setup::position(const string& name) const {
  map<string, HPT_ChamberID>::const_iterator iter;
  if ( (iter = mID.find(name)) == mID.end() ) return 0;
  else return &(*iter).second;
}

#endif
