// 
//  Author:  Rainer Mankel                     Created:  29-Sep-98
//           Humboldt University Berlin        Modified:  
//  Email:   mankel@ifh.de                     HERA-B Collaboration
// Modified:   
// - new flag NO_DEDX_CORR (undef)           25/07/00 A.Spiridonov
// - introduction of contructor from rseg
// - moved many routines from other state classes to here
//                                     14-Aug-00 Wouter Hulsbergen

#undef NO_DEDX_CORR

#include <iostream.h>
#include <math.h>
#include "rang/State.hh"
#include "rang/StateP.hh"
#include "rang/StatePI.hh"
#include "rang/StateL.hh"
#include "rang/StateLI.hh"
#include "rang/FitMode.hh"
#include "rang/ScatThick.hh"
#include "rang/zfpar.hh"
#include "rang/zfconst.hh"
#include "rang/RadEl.hh"
#include "rang/utfeld.hh"

//-- Constructor

rangerState::rangerState(int dim) : zState(0), qPEstimate(0.1), p(dim), c(dim) {}

rangerState::rangerState(const rangerState& s)  {
#ifndef FULL
  cout << "rangerState::rangerState: here is the copy constructor" << endl;
#endif
  zState     = s.zState;
  qPEstimate = s.qPEstimate;
  p          = s.p ;
  c          = s.c ;
#ifdef FULL
  cout << "rangerState::rangerState: Copy constructor done " << endl;
#endif
}

rangerState::~rangerState() {
  //  cout << "rangerState destructor has been called" << endl;
}

rangerState& rangerState::operator=(const rangerState& s) {
#ifdef FULL
   cout << "rangerState::operator=" << endl
	<< "   the given z is " << s.z() << endl;
#endif
   if(fitMode()==s.fitMode()) {
     zState     = s.zState;
     qPEstimate = s.qPEstimate;
     p          = s.p ;
     c          = s.c ;
   } else {
     cout << "rangerState::operator= called for two different derived states !" << endl ;
     exit(1) ;
   }
   return *this ;
}

void rangerState::covInit() {
  #ifdef FULL
  cout << "rangerStatePI::covInit: z=" << z() << endl;
  cout << "                     state matrix has " 
       << cov().num_row() << " " << cov().num_col() << endl;
#endif
  // reset
  cov() = HepSymMatrix(dim(),0) ;

  // fill
  setCovXX( zfpar_.cxxrf ) ;
  setCovYY( zfpar_.cyyrf ) ;
  setCovTxTx( zfpar_.ctxtxrf ) ;
  setCovTyTy( zfpar_.ctytyrf ) ;
  setCovQpQp( zfpar_.cqqrf ) ;
  setCovII( zfconst_.sigI * zfconst_.sigI ) ;

#ifdef FULL
  cout << "cov matrix is " << endl << c << endl;
  cout << "diag: " << zfpar_.cxxrf << " " << zfpar_.cyyrf << " " 
       << zfpar_.ctxtxrf << " " << zfpar_.ctytyrf << " "
       << zfpar_.cqqrf << endl ;
#endif
}

void
rangerState::initRHIT(const ArtePointer<RHIT>& rhit) 
{
  //Initialize with rhit
  zState = rhit->hitb->z;
  qPEstimate = rhit->p;
  setX(rhit->x) ;
  setY(rhit->y) ;
  setTx(rhit->tx) ;
  setTy(rhit->ty);
  setQp(rhit->p);
  setI(0) ;
}

void
rangerState::initHITB(const ArtePointer<HITB>& hitb) 
{
  //Initialize with hitb
  zState = hitb->z;
  qPEstimate = 0.1;
  setX(0) ;
  setY(0) ;
  setTx(0) ;
  setTy(0);
  setQp(0);
  setI(0) ;
}

void
rangerState::initRCCL(const ArtePointer<RCCL>& rccl, int charge) 
{
  //Initialize with hitb
  zState = rccl->z;
  qPEstimate = 1/rccl->esum;  // beware of charge ambiguity
  setX(rccl->x) ;
  setY(rccl->y) ;
  setTx(rccl->tx) ;
  setTy(rccl->ty) ;
  if (rccl->esum > 0) setQp(charge/rccl->esum) ;
  else                setQp(charge * 9999) ;
  setI(0) ;
}

#include <arte/RSEG_cmp.hh>
void 
rangerState::initRSEG(const ArtePointer<RSEG>& rseg, unsigned char side) 
{
  //Initialize with rseg
  zState = rseg->zf ;
  qPEstimate = 0.1 ;
  switch(rseg->cmp) 
    {
    case Rsegc::rich :
      zState = 1000.0 ; // According to Denis, this is what they assume in their corrections ...
      setX(0) ;
      setY(0) ;
      setQp(0) ;
      setTx(rseg->txf) ;
      setTy(rseg->tyf) ;
      break ;
    default:
      float *x, *c ;
      if(side==0) {
	x = &(rseg->xf) ;
	c = rseg->cvf ;
      } else {
	x = &(rseg->xe) ;
	c = rseg->cve ;
      }      
      setX(x[0]) ;
      setY(x[1]) ;
      setTx(x[2]) ;
      setTy(x[3]) ;
      setQp(x[4]) ;
      setI(rseg->time) ;
      
      // we could as well fill the covariance matrix
      setCovXX(  c[0]) ;
      setCovXY(  c[1]) ;
      setCovYY(  c[2]) ;
      setCovXTx( c[3]) ;
      setCovYTx( c[4]) ;
      setCovTxTx(c[5]) ;
      setCovXTy( c[6]) ;
      setCovYTy( c[7]) ;
      setCovTxTy(c[8]) ;
      setCovTyTy(c[9]) ;
      setCovXQp( c[10]) ;
      setCovYQp( c[11]) ;
      setCovTxQp(c[12]) ;
      setCovTyQp(c[13]) ;
      setCovQpQp(c[14]) ;
      setCovXI(0) ;
      setCovYI(0) ;
      setCovTxI(0) ;
      setCovTyI(0) ;
      setCovQpI(0) ;
      setCovII(rseg->dtim*rseg->dtim) ;
    }
}

rangerState* rangerState::makeState(const rangerFitMode fm) {
  rangerState* s;
  switch (fm) {
  case ModeP:
    s = new rangerStateP();
    break;
  case ModePI:
    s = new rangerStatePI();
    break;
  case ModeL:
    s = new rangerStateL();
    break;
  case ModeLI:
    s = new rangerStateLI();
    break;
  default:
    cout << "rangerState::makeState: improper mode " << int(fm) << endl;
    exit(1);
  }
  return s;
}

rangerState* rangerState::makeState( const ArtePointer<RHIT>& rhit,
				     const rangerFitMode fm ) 
{
  // Create state
  rangerState* s = makeState(fm) ;
  s->initRHIT(rhit) ;
  return s;
}


rangerState* rangerState::makeState( const ArtePointer<HITB>& hitb,
				     const rangerFitMode fm ) 
{
  // create state
  rangerState* s = makeState(fm) ;
  s->initHITB(hitb) ;
  return s ;
}

rangerState* rangerState::makeState( const ArtePointer<RCCL>& rccl,
				     int charge,const rangerFitMode fm ) {
  rangerState* s = makeState(fm) ;
  s->initRCCL(rccl,charge) ;
  return s ;
}

rangerState* rangerState::makeState( const ArtePointer<RSEG>& rseg, 
				     const rangerFitMode fm, unsigned char side) {
  rangerState* s = makeState(fm) ;
  s->initRSEG(rseg,side) ;
  return s ;
}

void rangerState::mergeRhit(ArtePointer<RHIT> rhit,float weight){
                                      // merge parameters to RHIT
  float w1 = 1 - weight;
  rhit->x  = w1 * rhit->x  + weight * x();
  rhit->y  = w1 * rhit->y  + weight * y();
  rhit->tx = w1 * rhit->tx + weight * tx();
  rhit->ty = w1 * rhit->ty + weight * ty();
  rhit->p  = w1 * rhit->p  + weight * qP();
}

void rangerState::copyVec( const rangerState& s) 
{
  zState = s.z();
  p = s.vec();
  qPEstimate = s.qPEst();
}

void rangerState::linExtrapVec( const rangerState& s, float z) 
{
  setZ( z ) ;
  setQpEst( s.qPEst() );
  setX( s.x() + (z - s.z()) * s.tx() );
  setY( s.y() + (z - s.z()) * s.ty() );
  setTx( s.tx() );
  setTy( s.ty() );
  setQp( s.qP() ) ;
  setI( s.I() ) ;
}

void rangerState::dump() {
  cout << "     zState=" << zState << "  vec=" << p << endl;
  cout << "                              cov=" << c << endl;
}

void rangerState::applyScat(const rangerScatThin& scat) {

  setCovTxTx( covTxTx() + scat.cTxTx ) ;
  setCovTyTy( covTyTy() + scat.cTyTy ) ;
  setCovTxTy( covTxTy() + scat.cTxTy ) ;

  float xField[3];
  xField[0] = x();
  xField[1] = y();
  xField[2] = zState;
  float hField[3];
  utfeld_(xField,hField);
  
  if (zState > zfpar_.zminMag && zState < zfpar_.zmaxMag) {
    if (fabs(hField[1]) > 0.1) {
      setCovQpQp( covQpQp() + zfpar_.fP * fabs(hField[1]) * scat.dScat );
    }
  }
}

void rangerState::applyScat(const rangerScatThick& scat) {

  setCovTxTx( covTxTx() + scat.cTxTx ) ;
  setCovTyTy( covTyTy() + scat.cTyTy ) ;
  setCovTxTy( covTxTy() + scat.cTxTy ) ;

  setCovXX( covXX() + scat.cXX ) ; // correction for thick scatterer
  setCovXY( covXY() + scat.cXY ) ;
  setCovYY( covYY() + scat.cYY ) ;
  setCovXTx( covXTx() + scat.cXTx ) ;
  setCovXTy( covXTy() + scat.cXTy ) ;
  setCovYTx( covYTx() + scat.cYTx ) ;
  setCovYTy( covYTy() + scat.cYTy ) ;

  float xField[3];
  xField[0] = x();
  xField[1] = y();
  xField[2] = zState;
  float hField[3];
  utfeld_(xField,hField);

  if (zState > zfpar_.zminMag && zState < zfpar_.zmaxMag) {
    if (fabs(hField[1]) > 0.1) {
      setCovQpQp( covQpQp() + zfpar_.fP * fabs(hField[1]) * scat.dScat );
    }
  }
}

void rangerState::applyRadEl(const rangerRadEl& rad) {
  if (zState > zfpar_.zminEc && zState < zfpar_.zmaxEc) {
    setQp( qP()*rad.qPFactor ) ;
    setCovQpQp( covQpQp() + rad.cQpQp ) ;
  }
}

void rangerState::applyDeDx(const float rl) {
  if (zState < zfpar_.zmaxMag && qP() != 0) {
    //float eLoss = 0.075 * rl;
#ifndef NO_DEDX_CORR
    //float eLoss = 0.075 * rl;
    float eLoss = 0.048 * rl;     // tuned for radThickEff
#else
    float eLoss;
    if( zfpar_.rlevel == 3 || zfpar_.rlevel == 4 ) 
      {
         eLoss = 0.048 * rl;     // tuned for radThickEff
      }
    else
      {
        eLoss = 0.;
      }
#endif
    setQp( (qP() > 0)? (1./(1./qP() + eLoss)) : (1./(1./qP() - eLoss)) );
  }
}

void rangerState::applyRadPhoton(const float e,const float dE) {
  if (qP() != 0) {
    double qPOld = qP();
    setQp( (qP() > 0)? (1./(1./qP() + e)) : (1./(1./qP() - e)) );
    double efac = qP() / qPOld;
    double dCov = (efac * efac * efac * efac - 1) * covQpQp()
      + dE * dE * qP() * qP() * qP() * qP();
    if (dCov >0) setCovQpQp( covQpQp() + dCov ) ;
  }
}

void rangerState::storeRseg(ArtePointer<RSEG> rseg,int firstLast){
 switch (firstLast) {
 case 1:
   rseg->zf = zState;
 
  rseg->xf = x();
   rseg->yf = y();
   rseg->txf = tx();
   rseg->tyf = ty();
   if(fabs(qP())>0) rseg->pf = qP();
   rseg->cvf[ 0] = covXX();
   rseg->cvf[ 1] = covXY();
   rseg->cvf[ 2] = covYY();
   rseg->cvf[ 3] = covXTx();
   rseg->cvf[ 4] = covYTx();
   rseg->cvf[ 5] = covTxTx();
   rseg->cvf[ 6] = covXTy();
   rseg->cvf[ 7] = covYTy();
   rseg->cvf[ 8] = covTxTy();
   rseg->cvf[ 9] = covTyTy();
   rseg->cvf[10] = covXQp();
   rseg->cvf[11] = covYQp();
   rseg->cvf[12] = covTxQp();
   rseg->cvf[13] = covTyQp();
   if(fabs(qP())>0) rseg->cvf[14] = covQpQp();
   rseg->time = I();
   rseg->dtim = covII()>0 ? sqrt(covII()) : 0;

   // just for monitoring: fill interaction time into hist
#ifdef rangerHIST
   if(mode()==modePI) {
     int id=129;
     float xxx = I() - zfconst_.tI;
     float yyy = 0;
     float www=1;
     hfill_(id,xxx,yyy,www);
     id=128;
     xxx = zfconst_.tI;
     yyy = I();
     hfill_(id,xxx,yyy,www);
     id=139;
     if ( cov()(6,6) <= 0 ) {
       cout << "StatePI::storeRseg: cov()(6,6)=" << cov()(6,6) << endl;
       return;
     }
     xxx = (I() - zfconst_.tI) / sqrt( cov()(6,6) );
     hfill_(id,xxx,yyy,www);
   }
#endif
   break;
 case 2:
   rseg->ze = zState;
   rseg->xe = x();
   rseg->ye = y();
   rseg->txe = tx();
   rseg->tye = ty();
   if(fabs(qP())>0) rseg->pe = qP();
   rseg->cve[ 0] = covXX();
   rseg->cve[ 1] = covXY();
   rseg->cve[ 2] = covYY();
   rseg->cve[ 3] = covXTx();
   rseg->cve[ 4] = covYTx();
   rseg->cve[ 5] = covTxTx();
   rseg->cve[ 6] = covXTy();
   rseg->cve[ 7] = covYTy();
   rseg->cve[ 8] = covTxTy();
   rseg->cve[ 9] = covTyTy();
   rseg->cve[10] = covXQp();
   rseg->cve[11] = covYQp();
   rseg->cve[12] = covTxQp();
   rseg->cve[13] = covTyQp();
   if(fabs(qP())>0) rseg->cve[14] = covQpQp();
   break;
 default:
   cout << "rangerStatePI::storeRseg: invalid firstLast " << firstLast << endl;
   exit(1);
   break;
 }
}

void rangerState::storeRtra(ArtePointer<RTRA> rtra,int firstLast){
 switch (firstLast) {
 case 1:
   rtra->zf = zState;
   rtra->xf = x();
   rtra->yf = y();
   rtra->txf = tx();
   rtra->tyf = ty();
   if(fabs(qP())>0) rtra->pf = qP();
   rtra->cvf[ 0] = covXX();
   rtra->cvf[ 1] = covXY();
   rtra->cvf[ 2] = covYY();
   rtra->cvf[ 3] = covXTx();
   rtra->cvf[ 4] = covYTx();
   rtra->cvf[ 5] = covTxTx();
   rtra->cvf[ 6] = covXTy();
   rtra->cvf[ 7] = covYTy();
   rtra->cvf[ 8] = covTxTy();
   rtra->cvf[ 9] = covTyTy();
   rtra->cvf[10] = covXQp();
   rtra->cvf[11] = covYQp();
   rtra->cvf[12] = covTxQp();
   rtra->cvf[13] = covTyQp();
   if(fabs(qP())>0) rtra->cvf[14] = covQpQp();
   break;
 case 2:
   break;
 default:
   break;
 }
}

void rangerState::storeRhit(ArtePointer<RHIT> rhit) {
  rhit->x = x();
  rhit->y = y();
  rhit->tx = tx();
  rhit->ty = ty();
  rhit->p = qP();
}

ArtePointer<RPNT> rangerState::storeRpnt(ArtePointer<RPNT> rpnt){
  ArtePointer<RPNT> rp;
  if (rpnt == 0) {
    // create new entry in RPNT. 
    mmbook_(ArtePointer<RPNT>::getLink().getRef(),1);
    rp = ArtePointer<RPNT>( ArtePointer<RPNT>::nrow() );
  }
  else {
    rp = rpnt;
  }
  //cout << "rangerState::storeRpnt: store in rpnt=" << rp << endl;
  if (rp != 0) {
    rp->zf = zState;
    rp->xf = x();
    rp->yf = y();
    rp->txf = tx();
    rp->tyf = ty();
    if(fabs(qP())>0) rp->pf = qP();
    rp->cvf[ 0] = covXX();
    rp->cvf[ 1] = covXY();
    rp->cvf[ 2] = covYY();
    rp->cvf[ 3] = covXTx();
    rp->cvf[ 4] = covYTx();
    rp->cvf[ 5] = covTxTx();
    rp->cvf[ 6] = covXTy();
    rp->cvf[ 7] = covYTy();
    rp->cvf[ 8] = covTxTy();
    rp->cvf[ 9] = covTyTy();
    rp->cvf[10] = covXQp();
    rp->cvf[11] = covYQp();
    rp->cvf[12] = covTxQp();
    rp->cvf[13] = covTyQp();
    if(fabs(qP())>0) rp->cvf[14] = covQpQp();
    return rp;
  }
  else {
    return 0;
  }
}


/*
void rangerState::linInterpolate( float z, const rangerState* s1, const rangerState* s2)
{
  // performs a linear _interpolation_ between states s1 and s2 to z
  // author: WDH

  setZ(z) ;
  if(s1->z()>s2->x()) swap(s1,s2) ;

  HepMatrix* F(0) ;

  if     ( z <= s1->z() ) s1->predict(z,this,F) ;
  else if( z >= s2->z() ) s2->predict(z,this,F) ;
  else {
    // extrapolate both states to z
    rangerState* es1(0),es2(0) ;
    s1->predict(z,es1,F) ;
    s2->predict(z,es2,F) ;
    
    // assign linear weight to both states
    float w1 = (s2->z() - z)/(s2->z() - s1->z()) ;
    float w2 = 1 - w1;
    
    // interpolate
    p = w1*es1->vec()    + w2*es2->vec() ;
    c = w1*w1*es1->cov() + w2*w2*es2->cov() ;
    setQpEst w1*es1
    delete es1,es2;
  }
  delete F;
}
*/
