// *****************************************************************************
//
// source:   Main reconstruction routine
//
// type:      source code
//
// created:   04 Aug 2000
//	
// author:    Denis Dujmic
//
// changes: 
//     10 Aug 2000 (DD) find geo acceptance (offline only)
//     15 Aug 2000 (DD) compare likelihoods (offline only)
//     07 Sep 2000 (DD) close ROOT files in destructor (offline only)
//     15 Sep 2000 (DD) refit of decays
//
// *****************************************************************************
#include "reco/rcX.h"
#include "reco/rcXkinema.h"
#include "reco/rcXsegment.h"
#include "reco/rcXdecay.h"

#ifdef USE_ROOT_DD
#include "mass/otr_util.h"
#endif

rcX::rcX(char *key, char *tit, int verbose): datamass(), vertexmass() {
  sprintf(keyword,"%s",key);
  sprintf(title,"%s",tit);
  total=0;
  old_evt=0;
  old_total=0;
  daugh_list= new vector<attribute>;
  first=true;
}

rcX::~rcX() { 
#ifdef  USE_ROOT_DD
  cout<<" Writing "<<keyword<<" file ... ";
  if (mother.root_output) x_file->Write();
  x_file->Close();
  cout<<"done"<<endl;
#endif
}

void rcX::bookXHistos() {

#ifdef  USE_ROOT_DD
  if (mother.root_output) x_file=createRootFile(keyword, title);

  x_loss   = new TH1F(  "loss",                 "Cuts",100, 0.5, 100.5);
  x_events = new TH1F("events","Events with mass reco",  3,-0.5,   2.5);
  x_fills  = new TH1F("nfill","Entries per event",  20,-0.5,   19.5);;
  
  x_tree = new TTree(keyword,"Mass reconstruction tree");
#endif

  const int nt=daugh_list->size();
  track    = new TRACK_PARAMETER[nt];
  mc_track = new TRACK_PARAMETER[nt];
  type     = new TYPE[nt];
  vds      = new TRACKER_SEGMENT[nt];
  itr      = new TRACKER_SEGMENT[nt];
  otr      = new TRACKER_SEGMENT[nt];
  rich     = new RICH_RING[nt];
  lr       = new LIKELIHOODS[nt];
  ecal     = new ECAL_CLUSTER[nt];
  brem     = new ECAL_CLUSTER[nt];
  muon     = new TRACKER_SEGMENT[nt];
  lr_rise  = new LIKELIHOODS[nt];
  lr_rite  = new LIKELIHOODS[nt];

#ifdef  USE_ROOT_DD
  char kw[10];
  int  i=1;
  for (vector<attribute>::iterator it=daugh_list->begin(); 
       it!=daugh_list->end(); ++it) {
    int ii=i-1;
    sprintf(kw,"%s%d",(*it).name,i); track[ii].branch(x_tree,kw);
    sprintf(kw,"vds%d",i);           vds[ii].branch  (x_tree,kw);
    sprintf(kw,"itr%d",i);           itr[ii].branch  (x_tree,kw);
    sprintf(kw,"otr%d",i);           otr[ii].branch  (x_tree,kw);
    sprintf(kw,"rich%d",i);          rich[ii].branch (x_tree,kw);
    sprintf(kw,"lr%d",i);            lr[ii].branch   (x_tree,kw);
    sprintf(kw,"ecal%d",i);          ecal[ii].branch (x_tree,kw);    
    if ((*daugh_list)[i].recoverBremsstrahlung) {
      sprintf(kw,"brem%d",i);        brem[ii].branch (x_tree,kw);
    }
    sprintf(kw,"muon%d",i);          muon[ii].branch (x_tree,kw);
    if (ArteTable<MTRA>::size()) {
      sprintf(kw,"type%d",i);             type[ii].branch    (x_tree,kw);
      sprintf(kw,"mc_%s%d",(*it).name,i); mc_track[ii].branch(x_tree,kw);
      sprintf(kw,"rise%d",i);             lr_rise[ii].branch (x_tree,kw);
      sprintf(kw,"rite%d",i);             lr_rite[ii].branch (x_tree,kw);
    }
    i++;
  }
  x_par.branch         (x_tree,"par");
  x_vertex.branch      (x_tree,"vertex");
  primary_vertex.branch(x_tree,"v0");
  x_wire.branch        (x_tree,"wire");
  event.branch         (x_tree,"event");
  if (ArteTable<MTRA>::size()) {
    mcx_vertex.branch       (x_tree,"mc_vertex");
    mc_primary_vertex.branch(x_tree,"mc_v0");
  }
#endif
}

void rcX::cleartree() {
  clear<TRACK_PARAMETER>(x_par);
  clear<VERTEX>(x_vertex);
  clear<VERTEX>(mcx_vertex);
  clear<VERTEX>(primary_vertex);
  clear<VERTEX>(mc_primary_vertex);
  clear<TARGET>(x_wire);
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    clear<TRACK_PARAMETER>(track[i]);
    clear<TRACKER_SEGMENT>(vds[i]);
    clear<TRACKER_SEGMENT>(itr[i]);
    clear<TRACKER_SEGMENT>(otr[i]);
    clear<RICH_RING>(rich[i]);
    clear<LIKELIHOODS>(lr[i]);
    clear<LIKELIHOODS>(lr_rise[i]);
    clear<LIKELIHOODS>(lr_rite[i]);
    clear<ECAL_CLUSTER>(ecal[i]);
    clear<TRACKER_SEGMENT>(muon[i]); 
    if (ArteTable<MTRA>::size()) {
      clear<TYPE>(type[i]);
    }
  }
}


void rcX::copy(int i) {
  segment::findSegments(*it_list[i]);
  track[i].copy(*it_list[i]);
  mc_track[i].copy((*(*it_list[i])->mtra.begin()));
  if (segment::vdsf())  { vds[i].copy(*segment::ivds);   } vds[i].hit1=(*it_list[i])->hsi; 
  if (segment::itrf())  { itr[i].copy(*segment::iitr);   } itr[i].hit1=(*it_list[i])->hin; 
  if (segment::otrf())  { otr[i].copy(*segment::iotr);   } otr[i].hit1=(*it_list[i])->hou; 
  if (segment::richf()) { rich[i].copy(*segment::irich); } rich[i].copy(*it_list[i]);

  lr[i].copy(*it_list[i]);
  if (segment::ecalf()) ecal[i].copy(*segment::iecal);
  if (segment::muonf()) { muon[i].copy(*segment::imuon); } muon[i].hit1=(*it_list[i])->hmu;
  if (segment::mtraf()) type[i].copy(*segment::imtra);

#ifdef USE_ROOT_DD
  lr_rise[i].copyRISE(*it_list[i]);
  lr_rite[i].copyRITE(*it_list[i]);

  // -- check otr acceptance, save result in otr.hit3
  double pou[6], pin[]={ (*it_list[i])->zf, (*it_list[i])->xf,(*it_list[i])->yf,
			   (*it_list[i])->txf,(*it_list[i])->tyf,(*it_list[i])->pf}; 
  kinema::swim(pin,pou,850);
  int notr_pl=otr_util::num_otr_planes(pou);
  otr[i].hit3=notr_pl;
  int nitr_pl=otr_util::num_itr_planes(pou);
  itr[i].hit3=nitr_pl;
#endif
}

void rcX::copytree() {
  for (unsigned int i=0; i<daugh_list->size(); i++) copy(i);
}


int rcX::massRefit(ArtePointer<RTRA> rtra) {
  cout<<"massRefit"<<endl;
  if (rtra->fit%100!=2 && rtra->fit%100!=3) return RCX_OK;
  if (!rtra->topp->lund)                    return RCX_MASS_CONSTRAINT_VERTEX;
  static vertexmass   rvtx;
  static vertexAttrib rattrib;
  rattrib.topp              = rtra->topp; // pointer to mother
  rattrib.rlevel            = 3;
  rattrib.change_track_rtra = 0;
  rattrib.max_vertex_chi2   = 7;
  rvtx.setVertex(rattrib);
  rvtx.removeAllVertexTracks();
  ArtePointer<RVER> vr=rtra->ever;
  for (ArteRelationIterator<RREL> irrel=vr->rrel.begin();
       irrel!=vr->rrel.end(); ++irrel) {
    rvtx.addVertexTrack((*irrel)->rtra, (*irrel)->rtra->topp);//////!!!!!!!! this won;t work - nobody is setting rtra-topp
  }
  int flag=rvtx.fitVertex();
  if (mother.verb) {
    cout<<"refit of decay topp="<<rtra->topp
	<<":  chi2="<<rvtx.vertexChi2()
	<<" flag="<<flag<<endl;
  }
  if (flag)                                 return RCX_MASS_CONSTRAINT_VERTEX;
  return RCX_OK;
}

int rcX::findTracksRTRA() {
  for (ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
       irtra!=ArteTable<RTRA>::end(); ++irtra) {
    // -- check daughter attributes and append track
    for (unsigned int i=0; i<daugh_list->size(); i++) {
      if (int ct=(*daugh_list)[i].checkAttributes((*irtra))) {
#ifdef  USE_ROOT_DD
	x_loss->Fill(ct);
#endif
	continue;
      }
      // -- check decays by refitting vertex with mass constraint
      if ((*daugh_list)[i].mass_refit && ((*irtra)->fit%100==2 || (*irtra)->fit%100==3)) {
	int ct=massRefit(*irtra);
	if (ct) {
#ifdef  USE_ROOT_DD
	  x_loss->Fill(ct);
#endif
	  continue;
	}
      }
      if ((*daugh_list)[i].useVdsOnce)
	rtra_all[i]->insert(pair<int, ArtePointer<RTRA> >((*irtra)->csvd,(*irtra))); 
      else
	rtra_list[i]->push_back((*irtra));
    }
  }

  // -- fill rtra_list with best chi2 for repeated VDS's
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    if (!(*daugh_list)[i].useVdsOnce) continue;
    it_rtra_all[i]=rtra_all[i]->begin();
    while (it_rtra_all[i]!=rtra_all[i]->end()) {
      ArtePointer<RTRA> best_rtra=(*it_rtra_all[i]).second; // rtra of 1st track
      int current=(*it_rtra_all[i]).first;                  // vds of 1st track
      for (it_rtra_all[i] =rtra_all[i]->lower_bound(current);  // all tracks with same vds 
	   it_rtra_all[i]!=rtra_all[i]->upper_bound(current);  //
	   ++it_rtra_all[i]) {                                 //
	if (best_rtra->chi>(*it_rtra_all[i]).second->chi)      //
	  best_rtra=(*it_rtra_all[i]).second;                  // choose best chi2
	else if (fabs(best_rtra->chi-(*it_rtra_all[i]).second->chi)<1e-2 &&
		 best_rtra->hsi+best_rtra->hin+best_rtra->hou<
		 (*it_rtra_all[i]).second->hsi+(*it_rtra_all[i]).second->hin+
		 (*it_rtra_all[i]).second->hou)
	  best_rtra=(*it_rtra_all[i]).second;                  // or choose more hits
      }                                                        //
      rtra_list[i]->push_back(best_rtra);                      // use best chi2
    }
  }
  return RCX_OK;
}


int  rcX::findRecoX() {
  if (!mother.fit)                             return RCX_MISSING_SEGMENTS;
  for (ArteTable<RTRA>::iterator irtra=ArteTable<RTRA>::begin();
       irtra!=ArteTable<RTRA>::end(); ++irtra) {
    if ((*irtra)->fit!=mother.fit) continue;
    ArtePointer<RVER> mothv=(*irtra)->ever;
    for (ArteRelationIterator<RREL> irrel=mothv->rrel.begin();
	 irrel!=mothv->rrel.end(); ++irrel) {
      ArtePointer<RTRA> rt=(*irrel)->rtra; 
      // -- check daughter attributes and append track
      for (unsigned int i=0; i<daugh_list->size(); i++) {
	if (int ct=(*daugh_list)[i].checkAttributes(rt)) {
#ifdef  USE_ROOT_DD
	  x_loss->Fill(ct);
#endif
	  continue;
	}
	rtra_list[i]->push_back(rt);
      }
    }
  }
  return RCX_OK;
}


int rcX::findTracksMTRA() {

  // -- check if decay valid
  if (mtraX()->leng<getVertAttrib().min_mtra_leng) return RCX_MC_CUT;
  float dz=mverX()->z - mverV0()->z;
  if (dz<getVertAttrib().min_dz_from_wire)         return RCX_MC_CUT;

  // -- find only mc tracks that belong to specific decay with existing rtra
  for (ArteTable<MTRA>::iterator imtra=ArteTable<MTRA>::begin();
       imtra!=ArteTable<MTRA>::end(); ++imtra) {
    // -- check if daughter
    int i=0;
    if ((*imtra)->rtra.begin()==(*imtra)->rtra.end()) continue;
    for (vector<attribute>::iterator it=daugh_list->begin();
	 it!=daugh_list->end(); ++it) {
      // -- take best chi2 / more hits
      ArtePointer<RTRA> best_rtra;
      float             best_chi=1e5;
      for (ArteRelationIterator<RTRA> irtra=(*imtra)->rtra.begin();
	     irtra!=(*imtra)->rtra.end(); ++irtra) {
//////////////// here use mc truth ///////////
	//float p=sqrt((*imtra)->px*(*imtra)->px + (*imtra)->py*(*imtra)->py + (*imtra)->pz*(*imtra)->pz);
	//(*irtra)->pf=1/p;
	//if ((*imtra)->topp->chrg<0) (*irtra)->pf= -(*irtra)->pf;
	//(*irtra)->txf=(*imtra)->px/(*imtra)->pz;
	//(*irtra)->tyf=(*imtra)->py/(*imtra)->pz;
	//(*irtra)->xf=(*imtra)->bver->x;
	//(*irtra)->yf=(*imtra)->bver->y;
	//(*irtra)->zf=(*imtra)->bver->z;
//////////////// here use mc truth ///////////
	if (int rc=(*it).checkAttributes(*irtra)) { 
	  //cout<<"mtra="<<(*imtra)<<"  loss="<<rc<<"  i="<<i<<endl;
#ifdef  USE_ROOT_DD
	  x_loss->Fill(rc); 
#endif
	  continue; 
	}
	// -- check decays by refitting vertex with mass constraint
	if ((*daugh_list)[i].mass_refit && ((*irtra)->fit%100==2 || (*irtra)->fit%100==3)) {
	  int ct=massRefit(*irtra);
	  if (ct) {
#ifdef  USE_ROOT_DD
	    x_loss->Fill(ct);
#endif
	    continue;
	  }
	}
	if ((*irtra)->chi<best_chi) { best_chi=(*irtra)->chi; best_rtra=(*irtra); }
      }
      if (best_chi<1e4) rtra_list[i]->push_back(best_rtra);
      i++;
    }
  }
  // -- check if all found
  unsigned int netra=0;
  for (ArteRelationIterator<MTRA> ietra=mtraX()->etra.begin();
       ietra!=mtraX()->etra.end(); ++ietra) netra++;
  if (netra!=daugh_list->size()) return RCX_MC_CUT;
  return RCX_OK;
}

void rcX::countLund() {
  if (!ArteTable<MTRA>::size()) return;
  unsigned int nd=0;
  // -- find mother
  for (ArteTable<MTRA>::iterator imtra=ArteTable<MTRA>::begin();
       imtra!=ArteTable<MTRA>::end(); ++imtra) {
    if ((*imtra)->topp==getVertAttrib().topp) { 
      if (mother.verb)
        cout<<"found mother="<<(*imtra)<<" lund="<<(*imtra)->topp->lund<<" grandmother="<<(*imtra)->btra<<endl;
      mtraX()=(*imtra); 
      break;
    }
  }

  // -- count proper decays
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    ArteTable<MTRA>::iterator imtra=ArteTable<MTRA>::begin();
    for (; imtra!=ArteTable<MTRA>::end(); ++imtra) {
      if ((*daugh_list)[i].lund==(*imtra)->topp->lund &&
	  ((*daugh_list)[i].mother==(*imtra)->btra->topp->lund||
	   !(*daugh_list)[i].mother)) break;
    }
    if (imtra!=ArteTable<MTRA>::end()) nd++;
  }

  // -- find mc vertices
  findVertexMC();
  unsigned int netra=0;
  for (ArteRelationIterator<MTRA> ietra=mtraX()->etra.begin();
       ietra!=mtraX()->etra.end(); ++ietra) netra++;
#ifdef  USE_ROOT_DD
  if (netra==nd && netra==daugh_list->size()) x_events->Fill(2);
#endif
}

void rcX::printTracks() {
  cout<<"rcX: Using tracks:"<<endl;
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    for (it_list[i]=rtra_list[i]->begin();
	 it_list[i]!=rtra_list[i]->end(); ++it_list[i]) cout<<" "<<(*it_list[i])<<" ";
    cout<<endl;
  }
  cout<<"total: ";
  for (unsigned int i=0; i<daugh_list->size(); i++) 
    cout<<" "<<rtra_list[i]->size()<<" ";
  cout<<"  =>  "<<total-old_total<<endl;
} 

void rcX::printDaughters() {
  cout<<"rcK: daughters:"<<endl;
  for (vector<attribute>::iterator it=daugh_list->begin();
       it!=daugh_list->end(); ++it) {
    cout<<" lund="<<(*it).lund<<"  mother="<<(*it).mother<<endl;
  }
  cout<<endl;
}


void rcX::rc() {

  // -- initialize
  if (first) {
    bookXHistos();
    for (unsigned int i=0; i<daugh_list->size(); i++) {
      rtra_list[i] = new vector<ArtePointer<RTRA> >;
      rtra_all[i]  = new multimap<int, ArtePointer<RTRA>, less<int> >;
    }
    first=false;
  }
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    if (rtra_list[i]->size()) rtra_list[i]->clear();
    if (rtra_all[i]->size())  rtra_all[i]->clear();
  }
  
  // -- count events
  event.copy();
#ifdef  USE_ROOT_DD
  x_events->Fill(0);
#endif
  countLund();

  // -- get tracks
  int cf=0;
  switch (mother.rlevel) {
  case 0: cf=findTracksRTRA(); break;
  case 1: cf=findRecoX();      break;
  case 2: cf=findTracksMTRA(); break;
  }
  if (cf) {
    if (mother.verb) printTracks();
    return;
  }
  for (unsigned int i=0; i<daugh_list->size(); i++) 
    if ((*daugh_list)[i].maxCandidate<(int)rtra_list[i]->size()) {
#ifdef  USE_ROOT_DD
      x_loss->Fill(RCX_MAX_TRACK_OCCUP);
#endif
      if (mother.verb) printTracks();
      return;
    }

  // -- make combination of tracks
  int ntyp=daugh_list->size();
  for (int i=0; i<ntyp; i++) {
    if (rtra_list[i]->begin()==rtra_list[i]->end())  return;
    if (rtra_list[i]->size()<(*daugh_list)[i].clone) return;
    it_list[i]=rtra_list[i]->begin()+(*daugh_list)[i].clone-1;
  }

  while (1) {
    for (int i=ntyp-1; i>0; i--) {
      if (it_list[i]==rtra_list[i]->end()) {
	it_list[i]=rtra_list[i]->begin()+(*daugh_list)[i].clone-1; 
	it_list[i-1]++;
      }
    }
    if (it_list[0]==rtra_list[0]->end()) break;

    // -- avoid multiple clone combinations
    int ii=1;
    for (; ii<ntyp-1; ii++) {
      if ((*daugh_list)[ii].clone<2) continue;
      if (it_list[ii]-rtra_list[ii]->begin()<=it_list[ii-1]-rtra_list[ii-1]->begin()) {
	it_list[ii]=  rtra_list[ii]->begin()+(it_list[ii-1]-rtra_list[ii-1]->begin()) + 1;
	break;
      }
    }
    if (ii<ntyp-1) continue;

    if (mother.verb) {
      cout<<"Combinations: ";
      for (unsigned int i=0; i<daugh_list->size(); i++) cout<<" "<<(*it_list[i])<<" ";
      cout<<endl;
    }

    cleartree();
    if (!TrackDiff() && !calcvertex() && !calcmass()) {
      copytree();
      bookX();
#ifdef  USE_ROOT_DD
      x_tree->Fill();
#endif
      // -- count
      if (old_evt!=ArtePointer<EVHD>(1)->evt) { 
	old_evt=ArtePointer<EVHD>(1)->evt;
#ifdef  USE_ROOT_DD
	x_events->Fill(1);
#endif
      }
      total++;
    }
    it_list[ntyp-1]++;
  }
#ifdef  USE_ROOT_DD
  x_fills->Fill(total-old_total);
#endif
  old_total=total;
  if (mother.verb) printTracks();
}


    
int rcX::calcmass() {
  if (!mother.calculate_mass) return RCX_OK;

  // -- find invariant mass of track combination
  HepLorentzVector x_4v;
  unsigned int i=0;
  for ( ; i<daugh_list->size(); i++) {
    if (fabs((*it_list[i])->pf)<1e-5) break; // check if momentum!=0
    HepLorentzVector t_4v=kinema::FourVector((*it_list[i]),(*daugh_list)[i].mass);
    x_4v += t_4v;
    if ((*daugh_list)[i].recoverBremsstrahlung) {
      float fbrem=kinema::findBremsstrahlung(*it_list[i]);
      brem[i].esum=fbrem;
      x_4v+=HepLorentzVector(t_4v.vect().unit()*fbrem, -fbrem);
    }
  }
  if (i!=daugh_list->size()) return RCX_P_ERROR1; // track(s) w/o momentum exist
  x_vertex.mass = sqrt(x_4v*x_4v);
  // -- check mass range
  if (x_vertex.mass>getVertAttrib().max_mass ||
      x_vertex.mass<getVertAttrib().min_mass) return RCX_MASS_CUT;
  x_vertex.p    = 1./x_4v.vect().mag();
  if (getVertAttrib().topp->chrg<0) x_vertex.p = -x_vertex.p;
  x_vertex.pt   = x_4v.vect().perp();
  x_vertex.tx   = x_4v.vect().x()/x_4v.vect().z();
  x_vertex.ty   = x_4v.vect().y()/x_4v.vect().z();
  x_par.tx = x_vertex.tx;
  x_par.ty = x_vertex.ty;
  x_par.ip = x_vertex.p;
  x_par.pt = x_vertex.pt;
  if (mother.verb) cout<<"x_4v="<<x_4v<<"  mass="<<x_vertex.mass<<endl;
  return RCX_OK;
}



int rcX::calcvertex() {

  // -- check if vertexing requested
  if (!getVertAttrib().rlevel) return RCX_OK;

  // -- fit vertex
  removeAllVertexTracks();
  ////////////////////cout<<"track combination for vertexing: "<<endl;
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    /////////////////cout<<" "<<(*it_list[i])<<" ";
    addVertexTrack((*it_list[i]),(*daugh_list)[i].topp);
  }
  if (int retf=fitVertex()) {
#ifdef  USE_ROOT_DD
    x_loss->Fill(retf); 
#endif
    return RCX_VERTEX_CUT; 
  }
  // -- fill tree variables
  x_vertex.chi=vertexChi2();
  x_vertex.x=vertexX(0);
  x_vertex.y=vertexX(1);
  x_vertex.z=vertexX(2);
  x_par.x   = x_vertex.x;
  x_par.y   = x_vertex.y;
  x_par.z   = x_vertex.z;
  x_vertex.ex = (vertexCS(0,0)>0) ? sqrt(vertexCS(0,0)) : 0;
  x_vertex.ey = (vertexCS(1,1)>0) ? sqrt(vertexCS(1,1)) : 0;
  x_vertex.ez = (vertexCS(2,2)>0) ? sqrt(vertexCS(2,2)) : 0;
  // -- fill used wire
  x_wire.copy(getDecayWire());
  // -- fill primary vertex
  primary_vertex.copy(rverV0());
  // -- fill MC vertex
  if (ArteTable<MTRA>::size()) { 
    // -- mc
    mcx_vertex.x=mverX()->x;
    mcx_vertex.y=mverX()->y;
    mcx_vertex.z=mverX()->z;
    float pt=mtraX()->px*mtraX()->px+mtraX()->py*mtraX()->py;
    mcx_vertex.pt=sqrt(pt);
    mcx_vertex.p=1./sqrt(pt+mtraX()->pz*mtraX()->pz);
    if (mtraX()->topp->chrg) mcx_vertex.p *= (mtraX()->topp->chrg/3.);
    mcx_vertex.tx=mtraX()->px/mtraX()->pz;
    mcx_vertex.ty=mtraX()->py/mtraX()->pz;
    mcx_vertex.L =mtraX()->leng;
    // -- primary
    mc_primary_vertex.x=mverV0()->x;
    mc_primary_vertex.y=mverV0()->y;
    mc_primary_vertex.z=mverV0()->z;
    if (mother.verb)
      cout<<"btra: "<<mtraX()<<" mver "<<mverX()<<": "<<mverX()->x<<" "<<mverX()->y<<" "<<mverX()->z<<endl;
  }

  return RCX_OK;
}


int rcX::TrackDiff() {
  for (unsigned int i=0; i<daugh_list->size(); i++) {
    for (unsigned int j=i+1; j<daugh_list->size(); j++) {
      if (segment::TrackDiff(*it_list[i],*it_list[j])) return RCX_TRACK_DIFF;
    }
  }
  return RCX_OK;
}


void rcX::bookRTRA() {
  RTRA rtra;
  memset(&rtra,0x0,sizeof(RTRA));
  rtra.xf  = x_vertex.x;
  rtra.yf  = x_vertex.y;
  rtra.zf  = x_vertex.z;
  rtra.txf = x_vertex.tx;
  rtra.tyf = x_vertex.ty;
  rtra.pf  = x_vertex.p;
  rtra.fit = mother.fit;
  rtra.chi = vertexChi2();
  memcpy(&rtra.cvf[0],&vertexCov(0),sizeof(float)*15);
  ArteTable<RTRA>::push_back(rtra);
  ArtePointer<RTRA> Rtra(ArteTable<RTRA>::size()); // rtra
  createRelation( getVertAttrib().topp, Rtra, &TOPP::rtra); // rtra-topp
  if (mother.rlevel==2)
    createRelation(      mtraX(), Rtra, &MTRA::rtra); // rtra-mtra
  if (getVertAttrib().rlevel) {
    ArtePointer<RVER>  ever(ArteTable<RVER>::size());
    Rtra->ever=ever; // rtra-rver
    ever->mass=x_vertex.mass;
    ever->btra=Rtra;
  }
}

void rcX::bookX() {
  if (!mother.book_tables) return;
  if (getVertAttrib().rlevel!=0 &&
      getVertAttrib().rlevel!=4)  bookRVER();
  bookRTRA(); 
}

