
// *****************************************************************************
//
// source:    
//
// type:      source code
//
// created:   26 Jul 2000
//	
// author:    Andreas Gellrich
//            HERA-B Collaboration
//            Humboldt-University Berlin
//            Physics Department
//            Invalidenstr. 110
//            10115 Berlin
//            Germany
//            E-mail: Andreas.Gellrich@desy.de
//
// changes:
// 26 Jul 2000 (AG) creation (code from Denis)
// 07 Aug 2000 (DD) electron selection (D.S.) 
// 05 Sep 2000 (DD) bug fix, always check rich likelihoods 
// 15 Sep 2000 (DD) refit of vertex flag
//
// *****************************************************************************

#include "reco/rcXattribute.h"
#include "reco/rcXkinema.h"

attribute::attribute() {

  /// Momentum cut 
  minPt=0,minP=-1e5,maxP=1e5;

  /// Track quality
  maxChi=1e5;

  /// VDS quality
  minNhits=minPhits=minHSI=-1000; 
  
  /// ITR/OTR quality
  minMClayers=minPClayers=minTClayers=minHOU=minHIN=minHIO=-1000;

  /// RICH particle id
  maxElectronR2P2=maxMuonR2P2=maxPionR2P2=maxKaonR2P2=maxProtonR2P2=1e5;
  minLRE=minLRMU=minLRPI=minLRK=minLRP=-1e5;
  /// RICH ring quality
  minRichRadius=0, maxRichRadius=1e5, minRichSN=0, minHRI=-1000;
  /// Exclude RICH bands
  maxLRE=maxLRMU=maxLRPI=maxLRK=maxLRP=1e5;


  /// ECAL quality
  maxEcalEnergy=1e5, minEcalWid=-1, minEoP=-1e5, maxEoP=1e5;
  minHCA=-1000;

  /// ECAL particle id
  minLEE=minLEMI=minLEH=-1e-5; 
  maxLEE=maxLEMI=maxLEH= 1e5;

  /// MUON quality
  minHMU=-1000;

  /// Combinations
  minHCAMU=minHRIOU=minHRIIN=-1000;

  /// MC truth
  lund=mother=0;
  topp=0;

  /// name
  sprintf(name,"t");

  /// fit type
  fit=0;

  /// mass range for reconstructed decays
  minMass=-1e5, maxMass=1e5, mass=0;

  /// Vertex Position
  minVertexZ=-1e5, maxVertexZ=1e5;

  /// Downstream vertex (not pointing to the target)
  min_dz_from_wire=-1e5;

  /// bremsstrahlung
  recoverBremsstrahlung=0;

  /// Use vds segment only once
  useVdsOnce=0;

  /// Clone
  clone=1;

  /// Maximum candidates
  maxCandidate=100000;

  /// Special algorithm to select electrons
  eDora=0;

  /// Special requirements to select muons
  muKen=0;

  /// Refit of vertex with mass constraint in case of a decay
  mass_refit=0;
}


int attribute::checkAttributes(ArtePointer<RTRA> track) {


  // -- charged track only
  if (track->fit%100!=fit)                                         return RCX_WRONG_FIT;

  // -- mass and vertex range for reconstructed decays
  if (track->fit%100==2 || track->fit%100==3) {
    if (track->ever->mass>maxMass  || track->ever->mass<minMass)   return RCX_MASS_CUT;
    if (track->ever->zv<minVertexZ || track->ever->zv>maxVertexZ)  return RCX_DECAY_Z_CUT; 
  }

  // -- momentum 
  float p=1/(track->pf);
  float pt=fabs(p)*sqrt(1-1/(track->txf*track->txf+track->tyf*track->tyf+1));
  if (p<minP)                         return RCX_P_MIN;
  if (p>maxP)                         return RCX_P_MAX;
  if (pt<minPt)                       return RCX_PT_MIN;

  // -- track components
  if (track->hsi<minHSI)              return RCX_MISSING_SEGMENTS;
  if (track->hin<minHIN)              return RCX_MISSING_SEGMENTS;
  if (track->hou<minHOU)              return RCX_MISSING_SEGMENTS;
  if (track->hin+track->hou<minHIO)   return RCX_MISSING_SEGMENTS;
  if (track->hri<minHRI)              return RCX_MISSING_SEGMENTS;
  if (track->hca<minHCA)              return RCX_MISSING_SEGMENTS;
  if (track->hmu<minHMU)              return RCX_MISSING_SEGMENTS;
  if (track->hca+track->hmu<minHCAMU) return RCX_MISSING_SEGMENTS;
  if (minHRIIN>0 && track->hin<=0 &&
      track->cer<0.01)                return RCX_MISSING_SEGMENTS; // require either ring or itr
  if (minHRIOU>0 && track->hou<=0 &&
      track->cer<0.01)                return RCX_MISSING_SEGMENTS; // require either ring or otr

  // -- RICH
  if (track->cer>0.01) {
    if (track->cer<minRichRadius)                             return RCX_RICH_MIN_RAD;
    if (track->cer>maxRichRadius)                             return RCX_RICH_MAX_RAD;
  }
  if (track->lre>maxElectronR2P2  || track->lre<minLRE)       return RCX_RICH_ID;
  if (track->lrmu>maxMuonR2P2     || track->lrmu<minLRMU)     return RCX_RICH_ID;
  if (track->lrpi>maxPionR2P2     || track->lrpi<minLRPI)     return RCX_RICH_ID;
  if (track->lrk>maxKaonR2P2      || track->lrk<minLRK)       return RCX_RICH_ID;
  if (track->lrp>maxProtonR2P2    || track->lrp<minLRP)       return RCX_RICH_ID;        
  if (track->lre>maxLRE)                                      return RCX_RICH_ID; // exclude bands
  if (track->lrmu>maxLRMU)                                    return RCX_RICH_ID; // exclude bands
  if (track->lrpi>maxLRPI)                                    return RCX_RICH_ID; // exclude bands
  if (track->lrk>maxLRK)                                      return RCX_RICH_ID; // exclude bands
  if (track->lrp>maxLRP)                                      return RCX_RICH_ID; // exclude bands


  // -- ECAL
  float eop=0;
  if (track->hca) {
    ArtePointer<RCCL> ecal=(*track->rccl.begin());
    eop=ecal->esum/fabs(p);
    if (ecal->wid<minEcalWid)                                   return RCX_ECAL_MIN_WID;
    if (ecal->esum>maxEcalEnergy)                               return RCX_ECAL_MAX_E;
    if (eop<minEoP)                                             return RCX_EOP_MIN;
    if (eop>maxEoP)                                             return RCX_EOP_MAX;
  }

  // -- electron selection cuts
  if (eDora) {
    if (track->hca<1)                                           return RCX_E_SELECT;
    if (eop<.7 || eop>1.3)                                      return RCX_E_SELECT;
    if (track->hin+track->hou<1 &&
	track->cer<(ArtePointer<RITH>(1)->rad0-.003) &&
	kinema::findBremsstrahlung(track)<1)                    return RCX_E_SELECT;
  }

  if (muKen) {
    //if (track->hmu<1)                                           return RCX_MU_SELECT;
    // ...
  }

  // -- MC
  if (track->mtra.begin()!=track->mtra.end()) {
    ArtePointer<MTRA> mtra=track->mtra.begin();
    if (lund &&  mtra->topp->lund!=lund)                        return RCX_MC_CUT;
    if (mother && mtra->btra->topp->lund!=mother)               return RCX_MC_CUT;
  }


  return RCX_OK;
}
